from __future__ import annotations
import json
import logging
from collections.abc import Mapping
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from nplinker.schemas import validate_podp_json
from nplinker.strain import StrainCollection
from .gnps.gnps_file_mapping_loader import GNPSFileMappingLoader
from .molecular_family import MolecularFamily
from .spectrum import Spectrum


logger = logging.getLogger(__name__)


def add_annotation_to_spectrum(
    annotations: Mapping[str, dict], spectra: Sequence[Spectrum]
) -> None:
    """Add annotations to the `Spectrum.gnps_annotations` attribute for input spectra.

    It is possible that some spectra don't have annotations.

    !!! note
        The input `spectra` list is changed in place.

    Args:
        annotations: A dictionary of GNPS annotations, where the keys are
            spectrum ids and the values are GNPS annotations.
        spectra: A list of Spectrum objects.
    """
    for spec in spectra:
        if spec.id in annotations:
            spec.gnps_annotations = annotations[spec.id]


def add_strains_to_spectrum(
    strains: StrainCollection, spectra: Sequence[Spectrum]
) -> tuple[list[Spectrum], list[Spectrum]]:
    """Add `Strain` objects to the `Spectrum.strains` attribute for input spectra.

    !!! note
        The input `spectra` list is changed in place.

    Args:
        strains: A collection of strain objects.
        spectra: A list of Spectrum objects.

    Returns:
        A tuple of two lists of Spectrum objects,

            - the first list contains Spectrum objects that are updated with Strain objects;
            - the second list contains Spectrum objects that are not updated with Strain objects
            because no Strain objects are found.
    """
    spectra_with_strains = []
    spectra_without_strains = []
    for spec in spectra:
        try:
            strain_list = strains.lookup(spec.id)
        except ValueError:
            spectra_without_strains.append(spec)
            continue

        for strain in strain_list:
            spec.strains.add(strain)
        spectra_with_strains.append(spec)

    logger.info(
        f"{len(spectra_with_strains)} Spectrum objects updated with Strain objects.\n"
        f"{len(spectra_without_strains)} Spectrum objects not updated with Strain objects."
    )

    return spectra_with_strains, spectra_without_strains


def add_spectrum_to_mf(
    spectra: Sequence[Spectrum], mfs: Sequence[MolecularFamily]
) -> tuple[list[MolecularFamily], list[MolecularFamily], dict[MolecularFamily, set[str]]]:
    """Add Spectrum objects to MolecularFamily objects.

    The attribute `MolecularFamily.spectra_ids` contains the ids of `Spectrum` objects.
    These ids are used to find `Spectrum` objects from the input `spectra` list. The found `Spectrum`
    objects are added to the `MolecularFamily.spectra` attribute.

    It is possible that some spectrum ids are not found in the input `spectra` list, and so their
    `Spectrum` objects are missing in the `MolecularFamily` object.


    !!! note
        The input `mfs` list is changed in place.

    Args:
        spectra: A list of Spectrum objects.
        mfs: A list of MolecularFamily objects.

    Returns:
        A tuple of three elements,

            - the first list contains `MolecularFamily` objects that are updated with `Spectrum` objects
            - the second list contains `MolecularFamily` objects that are not updated with `Spectrum`
            objects (all `Spectrum` objects are missing).
            - the third is a dictionary containing `MolecularFamily` objects as keys and a set of ids
            of missing `Spectrum` objects as values.
    """
    spec_dict = {spec.id: spec for spec in spectra}
    mf_with_spec = []
    mf_without_spec = []
    mf_missing_spec: dict[MolecularFamily, set[str]] = {}
    for mf in mfs:
        for spec_id in mf.spectra_ids:
            try:
                spec = spec_dict[spec_id]
            except KeyError:
                if mf not in mf_missing_spec:
                    mf_missing_spec[mf] = {spec_id}
                else:
                    mf_missing_spec[mf].add(spec_id)
                continue
            mf.add_spectrum(spec)

        if mf.spectra:
            mf_with_spec.append(mf)
        else:
            mf_without_spec.append(mf)

    logger.info(
        f"{len(mf_with_spec)} MolecularFamily objects updated with Spectrum objects.\n"
        f"{len(mf_without_spec)} MolecularFamily objects not updated with Spectrum objects.\n"
        f"{len(mf_missing_spec)} MolecularFamily objects have missing Spectrum objects."
    )
    return mf_with_spec, mf_without_spec, mf_missing_spec


# ------------------------------------------------------------------------------
# Functions to extract mappings for metabolomics side:
# strain_id <-> MS_filename <-> spectrum_id
# ------------------------------------------------------------------------------
def extract_mappings_strain_id_ms_filename(
    podp_project_json_file: str | PathLike,
) -> dict[str, set[str]]:
    """Extract mappings "strain_id <-> MS_filename".

    Args:
        podp_project_json_file: The path to the PODP project JSON file.

    Returns:
        Key is strain id and value is a set of MS filenames.

    Notes:
        The `podp_project_json_file` is the project JSON file downloaded from
        PODP platform. For example, for project MSV000079284, its json file is
        https://pairedomicsdata.bioinformatics.nl/api/projects/4b29ddc3-26d0-40d7-80c5-44fb6631dbf9.4.

    See Also:
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    mappings_dict: dict[str, set[str]] = {}
    with open(podp_project_json_file, "r") as f:
        json_data = json.load(f)

    validate_podp_json(json_data)

    # Extract mappings strain id <-> metabolomics filename
    for record in json_data["genome_metabolome_links"]:
        strain_id = record["genome_label"]
        # get the actual filename of the mzXML URL
        filename = Path(record["metabolomics_file"]).name
        if strain_id in mappings_dict:
            mappings_dict[strain_id].add(filename)
        else:
            mappings_dict[strain_id] = {filename}
    return mappings_dict


def extract_mappings_ms_filename_spectrum_id(
    gnps_file_mappings_file: str | PathLike,
) -> dict[str, set[str]]:
    """Extract mappings "MS_filename <-> spectrum_id".

    Args:
        gnps_file_mappings_file: The path to the GNPS file mappings file (csv or tsv).

    Returns:
        Key is MS filename and value is a set of spectrum ids.

    Notes:
        The `gnps_file_mappings_file` is downloaded from GNPS website and named as
        [GNPS_FILE_MAPPINGS_TSV][nplinker.defaults.GNPS_FILE_MAPPINGS_TSV] or
        [GNPS_FILE_MAPPINGS_CSV][nplinker.defaults.GNPS_FILE_MAPPINGS_CSV].
        For more details, see [GNPS data][gnps-data].

    See Also:
        - [GNPSFileMappingLoader][nplinker.metabolomics.gnps.gnps_file_mapping_loader.GNPSFileMappingLoader]:
        Load GNPS file mappings file.
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    loader = GNPSFileMappingLoader(gnps_file_mappings_file)
    return loader.mapping_reversed


def get_mappings_strain_id_spectrum_id(
    mappings_strain_id_ms_filename: Mapping[str, set[str]],
    mappings_ms_filename_spectrum_id: Mapping[str, set[str]],
) -> dict[str, set[str]]:
    """Get mappings "strain_id <-> spectrum_id".

    Args:
        mappings_strain_id_ms_filename: Mappings
            "strain_id <-> MS_filename".
        mappings_ms_filename_spectrum_id: Mappings
            "MS_filename <-> spectrum_id".

    Returns:
        Key is strain id and value is a set of spectrum ids.


    See Also:
        - `extract_mappings_strain_id_ms_filename`: Extract mappings "strain_id <-> MS_filename".
        - `extract_mappings_ms_filename_spectrum_id`: Extract mappings "MS_filename <-> spectrum_id".
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    mappings_dict = {}
    for strain_id, ms_filenames in mappings_strain_id_ms_filename.items():
        spectrum_ids = set()
        for ms_filename in ms_filenames:
            if (sid := mappings_ms_filename_spectrum_id.get(ms_filename)) is not None:
                spectrum_ids.update(sid)
        if spectrum_ids:
            mappings_dict[strain_id] = spectrum_ids
    return mappings_dict



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import multiprocessing
import os
import time
import iokr_opt
import iokrdata as iokrdataserver
import numpy


def normalise_kernel(matrix):
    return matrix / numpy.sqrt(numpy.outer(matrix.diagonal(), matrix.diagonal()))


def load_kernel_file(filename):
    kernel = numpy.load(filename)
    return normalise_kernel(kernel)


def load_kernels(kernel_files):
    kernel_matrices = [load_kernel_file(x) for x in kernel_files]
    kernel_sum = numpy.sum(kernel_matrices, axis=0)
    return normalise_kernel(kernel_sum)


def gather_results(active_jobs, limit=25):
    done_jobs = []
    while len(active_jobs) > limit:
        remaining_jobs = []
        for line in active_jobs:
            job = line[-1]
            if job.ready():
                res = job.get()
                new_line = list(line[:-1])
                new_line.append(res)
                done_jobs.append(new_line)
            else:
                remaining_jobs.append(line)
        active_jobs = remaining_jobs
        time.sleep(1)
    return active_jobs, done_jobs


def run_iokr(data):
    p = multiprocessing.Pool()
    active_jobs = []
    job_limit = 25

    excluded = []
    missing_candidate = []
    collected_rankings = []

    for label in sorted(list(set(data.folds))):
        print("label %s" % label)
        label_indices = data.get_indices(label, complement=True)

        iokr = iokr_opt.InputOutputKernelRegression(data)
        iokr.set_training_indices(label_indices, _lambda=0.001)
        iokr.fit()

        test_indices = data.get_indices(label)
        for i in test_indices:
            if i not in data.test_sample_indices:
                excluded.append(i)
                continue

            sample = data.get_sample(i)
            formula = sample["formula"]
            sample_inchi = sample["inchi"]

            candidates = data.get_candidates(formula)
            candidate_inchi = [x[0] for x in candidates]
            correct_index = candidate_inchi.index(sample_inchi)

            # TODO: Recalculate fingerprints
            candidates = data.get_candidates(formula)
            candidate_fingerprints = [numpy.array(x[1]) for x in candidates]

            total_count = len(candidate_inchi)

            print(f"iokr job idx {i}, cand.set size {total_count}")
            # ranking = iokr.rank_candidates_opt(i, candidate_fingerprints)
            # # print(ranking)
            # ranking = list(ranking)
            # correct_ranking = ranking.index(correct_index)
            # print('ranked {} / {}'.format(correct_ranking, total_count))

            latent, x_kernel_vector, latent_basis, gamma = iokr.get_data_for_candidate_ranking(i)
            args = (i, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma)
            job = p.apply_async(iokr_opt.rank_candidates_opt, args)
            active_jobs.append((i, formula, correct_index, label, total_count, job))

            if len(active_jobs) > job_limit:
                active_jobs, results = gather_results(active_jobs, job_limit)
                for (
                    res_i,
                    res_formula,
                    res_correct_index,
                    res_label,
                    res_total_count,
                    res_output,
                ) in results:
                    res_ranking = list(res_output[0])
                    # print(res_ranking)
                    correct_ranking = res_ranking.index(res_correct_index)

                    collected_rankings.append((res_i, correct_ranking, res_total_count))

                    total = len(collected_rankings)
                    print(float([x[1] for x in collected_rankings].count(0)) / total, total)

                    # print(cr_b[res_i], cr_a[res_i], cr_a[res_i] == cr_b[res_i])

    print("Clean up remaining jobs")

    # clean up the last remaining jobs
    active_jobs, results = gather_results(active_jobs, 0)
    for res_i, res_formula, res_correct_index, res_label, res_total_count, res_output in results:
        res_ranking = list(res_output[0])
        correct_ranking = res_ranking.index(res_correct_index)
        collected_rankings.append((res_i, correct_ranking, res_total_count))
        total = len(collected_rankings)
        print(float([x[1] for x in collected_rankings].count(0)) / total, total)

    print("")
    print("IOKR test run done!")
    print(f"#samples: {len(collected_rankings)}")
    print(
        "top-1 acc: {}".format(
            float([x[1] for x in collected_rankings].count(0)) / total,
        )
    )

    return collected_rankings


def main():
    parser = argparse.ArgumentParser("Run IOKR test on a set")
    parser.add_argument("--kernel", dest="kernel", help="Kernel files", nargs="+")
    parser.add_argument(
        "--fp",
        dest="fingerprint",
        help="fingerprint type (substructure, cdk (default), klekota-roth",
        default="cdk_default",
    )
    parser.add_argument("--data", dest="datapath", help="data path", required=True)
    parser.add_argument("--output", dest="output", help="output label", required=True)
    args = parser.parse_args()

    # read from args
    # datapath = '/home/grimur/iokr/data'
    datapath = args.datapath
    # kernel_files = [datapath + os.sep + 'input_kernels_gh/ppk_dag_all_normalised_shifted_nloss.npy',
    #                 datapath + os.sep + 'input_kernels_gh/ppk_dag_all_normalised_shifted_peaks.npy']
    kernel_files = args.kernel
    # fingerprint = None

    fingerprint = args.fingerprint

    output_file = "IOKRranking_%s.bin" % args.output

    iokrdata = iokrdataserver.IOKRDataServer(datapath, kernel=None)
    kernel_matrix = load_kernels(kernel_files)
    iokrdata.kernel = kernel_matrix
    with open(datapath + os.sep + "ind_eval.txt") as f:
        raw_data = f.read()
        test_sample_indices = [int(x) - 1 for x in raw_data.strip().split()]
    iokrdata.test_sample_indices = test_sample_indices

    if fingerprint is not None:
        iokrdata.set_fingerprint(fingerprint)

    print("run iokr")
    rankings = run_iokr(iokrdata)

    numpy.save(output_file, rankings)


if __name__ == "__main__":
    main()



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pickle
import numpy

# import sys
# sys.path.append('/home/grimur/git/lda')
# from lda.code.formula import Formula
from .formula import Formula


global _ALL_DAG_MASSES
global _ALL_TREE_MASSES
_ALL_DAG_MASSES = None
_ALL_TREE_MASSES = None

global datapath
datapath = ""


def load_formula_dag(filename):
    formula = set()

    with open(filename) as f:
        for line in f.readlines():
            if not line.startswith("v"):
                continue

            fields = line.strip().split()
            if fields[1] == "->":
                # network edge
                f = fields[3].split('="')[1][:-3]
            else:
                # network vertex
                f = line.split("\\n")[0].split('"')[-1]

            formula.add(f)

    return list(formula)


def load_formula_tree(filename):
    with open(filename, "rb") as f:
        data = json.load(f)

    formula = [data["molecularFormula"], data["root"]]
    formula.extend([x["molecularFormula"] for x in data["fragments"]])
    formula.extend([x["source"] for x in data["losses"]])
    formula.extend([x["target"] for x in data["losses"]])
    formula.extend([x["molecularFormula"] for x in data["losses"]])

    return list(set(formula))


def load_peaks_from_tree(filename):
    with open(filename, "rb") as f:
        data = json.load(f)

    spectrum = []
    for f in data["fragments"]:
        if len(f["peaks"]) > 0:
            for p in f["peaks"]:
                mass = p["mz"]
                intensity = p["int"]
                spectrum.append((mass, intensity))
        else:
            mass = f["mz"]
            intensity = f["intensity"]
            spectrum.append((mass, intensity))

    return numpy.array(spectrum)


# These functions should be hanged onto the MS object
def filter_by_tree(self):
    treepath = "/home/grimur/iokr/data/trees"
    tol = 0.005

    try:
        formula = load_formula_tree(treepath + os.path.sep + self.id + ".json")
    except:
        formula = load_formula_dag(treepath + os.path.sep + self.id + ".dot")
    formula_objects = [Formula(x) for x in formula]
    formula_masses = sorted(x.compute_exact_mass() for x in formula_objects)

    return filter_by_mass(self.shifted_spectrum, formula_masses, tol)


def filter_by_tree_unshifted(self):
    treepath = "/home/grimur/iokr/data/trees"

    try:
        spectrum = load_peaks_from_tree(treepath + os.path.sep + self.id + ".json")
    except FileNotFoundError:
        spectrum = self.raw_spectrum

    return spectrum


def filter_by_dag(self):
    treepath = "/home/grimur/iokr/data/trees"
    tol = 0.005

    formula = load_formula_dag(treepath + os.sep + self.id + ".dot")
    formula_objects = [Formula(x) for x in formula]
    formula_masses = sorted(x.compute_exact_mass() for x in formula_objects)

    return filter_by_mass(self.shifted_spectrum, formula_masses, tol)


def load_all_dag_masses(path):
    formula_masses_collected = []
    for filename in os.listdir(path):
        if filename.endswith(".dot"):
            formula = load_formula_dag(path + os.sep + filename)
            formula_objects = [Formula(x) for x in formula]
            formula_masses_collected.extend([x.compute_exact_mass() for x in formula_objects])

    return sorted(list(set(formula_masses_collected)))


def filter_by_collected_dag(self):
    treepath = "/home/grimur/iokr/data/trees"
    tol = 0.005
    global _ALL_DAG_MASSES

    if _ALL_DAG_MASSES is None:
        _ALL_DAG_MASSES = load_all_dag_masses(treepath)

    return filter_by_mass(self.shifted_spectrum, _ALL_DAG_MASSES, tol)


def filter_by_frozen_dag(self):
    with open(os.path.join(datapath, "dag_masses.bin"), "rb") as f:
        dag_masses = pickle.load(f)
    tol = 0.005
    return filter_by_mass(self.shifted_spectrum, dag_masses, tol)


def load_all_tree_masses(path):
    formula_masses_collected = []
    for filename in os.listdir(path):
        if filename.endswith(".json"):
            formula = load_formula_tree(path + os.sep + filename)
            formula_objects = [Formula(x) for x in formula]
            formula_masses_collected.extend([x.compute_exact_mass() for x in formula_objects])

    return sorted(list(set(formula_masses_collected)))


def filter_by_collected_tree(self):
    treepath = "/home/grimur/iokr/data/trees"
    tol = 0.005
    global _ALL_TREE_MASSES

    if _ALL_TREE_MASSES is None:
        _ALL_TREE_MASSES = load_all_tree_masses(treepath)

    return filter_by_mass(self.shifted_spectrum, _ALL_TREE_MASSES, tol)


def filter_by_mass(raw_spectrum, formula_masses, tol):
    filtered_spectrum = []
    spectrum_index = 0
    mass_index = 0
    while mass_index < len(formula_masses):
        while (
            spectrum_index < len(raw_spectrum)
            and raw_spectrum[spectrum_index, 0] < formula_masses[mass_index] + tol
        ):
            while (
                spectrum_index < len(raw_spectrum)
                and raw_spectrum[spectrum_index, 0] < formula_masses[mass_index] - tol
            ):
                spectrum_index += 1
            if (
                spectrum_index < len(raw_spectrum)
                and raw_spectrum[spectrum_index, 0] < formula_masses[mass_index] + tol
            ):
                filtered_spectrum.append(raw_spectrum[spectrum_index, :])
            spectrum_index += 1
        mass_index += 1

    return numpy.array(filtered_spectrum)



from __future__ import annotations
import logging
import os
from deprecated import deprecated
from dynaconf import Dynaconf
from nplinker import defaults
from nplinker.defaults import NPLINKER_APP_DATA_DIR
from nplinker.genomics import BGC
from nplinker.genomics import GCF
from nplinker.genomics.antismash import AntismashBGCLoader
from nplinker.genomics.bigscape import BigscapeGCFLoader
from nplinker.genomics.bigscape import BigscapeV2GCFLoader
from nplinker.genomics.mibig import MibigLoader
from nplinker.genomics.utils import add_bgc_to_gcf
from nplinker.genomics.utils import add_strain_to_bgc
from nplinker.genomics.utils import get_mibig_from_gcf
from nplinker.metabolomics import MolecularFamily
from nplinker.metabolomics import Spectrum
from nplinker.metabolomics.gnps import GNPSAnnotationLoader
from nplinker.metabolomics.gnps import GNPSMolecularFamilyLoader
from nplinker.metabolomics.gnps import GNPSSpectrumLoader
from nplinker.metabolomics.utils import add_annotation_to_spectrum
from nplinker.metabolomics.utils import add_spectrum_to_mf
from nplinker.metabolomics.utils import add_strains_to_spectrum
from nplinker.strain import StrainCollection
from nplinker.strain.utils import load_user_strains


logger = logging.getLogger(__name__)


class DatasetLoader:
    """Load datasets from the working directory with the given configuration.

    ??? info "Concept and Diagram"
        [Working Directory Structure][working-directory-structure]

        [Dataset Loading Pipeline][dataset-loading-pipeline]

    Loaded data are stored in the data containers (attributes), e.g. `self.bgcs`, `self.gcfs`, etc.

    Attributes:
        config: A Dynaconf object that contains the configuration settings.
        bgcs: A list of BGC objects.
        gcfs: A list of GCF objects.
        spectra: A list of Spectrum objects.
        mfs: A list of MolecularFamily objects.
        mibig_bgcs: A list of MIBiG BGC objects.
        mibig_strains_in_use: A StrainCollection object that contains the strains in use from MIBiG.
        product_types: A list of product types.
        strains: A StrainCollection object that contains all strains.
        class_matches: A ClassMatches object that contains class match info.
        chem_classes: A ChemClassPredictions object that contains chemical class predictions.
    """

    RUN_CANOPUS_DEFAULT = False
    EXTRA_CANOPUS_PARAMS_DEFAULT = "--maxmz 600 formula zodiac structure canopus"

    # class predictions
    OR_CANOPUS = "canopus_dir"
    OR_MOLNETENHANCER = "molnetenhancer_dir"

    def __init__(self, config: Dynaconf) -> None:
        """Initialize the DatasetLoader.

        Args:
            config: A Dynaconf object that contains the configuration settings.

        Examples:
            >>> from nplinker.config import load_config
            >>> from nplinker.loader import DatasetLoader
            >>> config = load_config("nplinker.toml")
            >>> loader = DatasetLoader(config)
            >>> loader.load()

        See Also:
            [DatasetArranger][nplinker.arranger.DatasetArranger]: Download, generate and/or validate
                datasets to ensure they are ready for loading.
        """
        self.config = config

        self.bgcs: list[BGC] = []
        self.gcfs: list[GCF] = []
        self.spectra: list[Spectrum] = []
        self.mfs: list[MolecularFamily] = []
        self.mibig_bgcs: list[BGC] = []
        self.mibig_strains_in_use: StrainCollection = StrainCollection()
        self.product_types: list = []
        self.strains: StrainCollection = StrainCollection()

        self.class_matches = None
        self.chem_classes = None

    def load(self) -> bool:
        """Load all data from data files in the working directory.

        See [Dataset Loading Pipeline][dataset-loading-pipeline] for the detailed steps.

        Returns:
            True if all data are loaded successfully.
        """
        if not self._load_strain_mappings():
            return False

        if not self._load_metabolomics():
            return False

        if not self._load_genomics():
            return False

        # set self.strains with all strains from input plus mibig strains in use
        self.strains = self.strains + self.mibig_strains_in_use

        if len(self.strains) == 0:
            raise Exception("Failed to find *ANY* strains.")

        return True

    def _load_strain_mappings(self):
        # 1. load strain mappings
        sc = StrainCollection.read_json(self.config.root_dir / defaults.STRAIN_MAPPINGS_FILENAME)
        for strain in sc:
            self.strains.add(strain)
        logger.info("Loaded {} non-MiBIG Strain objects".format(len(self.strains)))

        # 2. filter user specified strains (remove all that are not specified by user).
        # It's not allowed to specify empty list of strains, otherwise validation will fail.
        user_strains_file = self.config.root_dir / defaults.STRAINS_SELECTED_FILENAME
        if user_strains_file.exists():
            logger.info(f"Loading user specified strains from file {user_strains_file}.")
            user_strains = load_user_strains(user_strains_file)
            logger.info(f"Loaded {len(user_strains)} user specified strains.")
            self.strains.filter(user_strains)

        logger.info("Loaded {} Strain objects in total".format(len(self.strains)))
        return True

    def _load_metabolomics(self):
        """Loads metabolomics data to Spectrum and MolecularFamily objects.

        The attribute of `self.spectra` is set to the loaded Spectrum objects that have Strain
        objects added (i.e. `Spectrum.strains` updated). If a Spectrum object does not have Strain
        objects, it is not added to `self.spectra`.

        The attribute of `self.mfs` is set to the loaded MolecularFamily objects that have
        Strain objects added (i.e. `MolecularFamily._strains` updated). This means only Spectra
        objects with updated strains (i.e. `self.spectra`) can be added to MolecularFamily objects.
        """
        logger.info(f"{'='*40}\nLoading metabolomics data starts...")

        gnps_dir = self.config.root_dir / defaults.GNPS_DIRNAME

        # Step 1: load all Spectrum objects
        raw_spectra = GNPSSpectrumLoader(gnps_dir / defaults.GNPS_SPECTRA_FILENAME).spectra
        # Step 2: load all GNPS annotations
        raw_annotations = GNPSAnnotationLoader(
            gnps_dir / defaults.GNPS_ANNOTATIONS_FILENAME
        ).annotations
        # Step 3: load all MolecularFamily objects
        raw_mfs = GNPSMolecularFamilyLoader(
            gnps_dir / defaults.GNPS_MOLECULAR_FAMILY_FILENAME
        ).get_mfs(keep_singleton=False)

        # Step 4: add GNPS annotations to Spectrum.gnps_annotations
        add_annotation_to_spectrum(raw_annotations, raw_spectra)
        # Step 5: add strains to Spectrum.strains
        spectra_with_strains, _ = add_strains_to_spectrum(self.strains, raw_spectra)

        # Step 6: add Spectrum objects to MolecularFamily
        mf_with_spec, _, _ = add_spectrum_to_mf(spectra_with_strains, raw_mfs)

        # Step 7: set attributes of self.spectra and self.mfs with valid objects
        self.spectra = spectra_with_strains
        self.mfs = mf_with_spec

        logger.info("Loading metabolomics data completed\n")
        return True

    def _load_genomics(self):
        """Loads genomics data to BGC and GCF objects.

        The attribute of `self.bgcs` is set to the loaded BGC objects that have the Strain object
        added (i.e. `BGC.strain` updated). If a BGC object does not have the Strain object, it is
        not added to `self.bgcs`. For MIBiG BGC objects, only those in use are added to `self.bgcs`.

        The attribute of `self.gcfs` is set to the loaded GCF objects that have the Strain objects
        added (i.e. `GCF._strains` updated). This means only BGC objects with updated Strain objects
        (i.e. `self.bgcs`) can be added to GCF objects.
        """
        logger.info(f"{'='*40}\nLoading genomics data starts...")

        # Step 1: load antismash BGC objects & add strain info
        logger.info("Parsing AntiSMASH directory...")
        antismash_bgcs = AntismashBGCLoader(
            str(self.config.root_dir / defaults.ANTISMASH_DIRNAME)
        ).get_bgcs()
        antismash_bgcs_with_strain, _ = add_strain_to_bgc(self.strains, antismash_bgcs)

        # Step 2: load mibig BGC objects (having strain info)
        if self.config.mibig.to_use:
            self.mibig_bgcs = MibigLoader(
                str(self.config.root_dir / defaults.MIBIG_DIRNAME)
            ).get_bgcs()

        # Step 3: get all BGC objects with strain info
        all_bgcs_with_strain = antismash_bgcs_with_strain + self.mibig_bgcs

        # Step 4: load all GCF objects
        if self.config.bigscape.version == 1:
            bigscape_cluster_file = (
                self.config.root_dir
                / defaults.BIGSCAPE_DIRNAME
                / f"mix_clustering_c{self.config.bigscape.cutoff}.tsv"
            )
            loader = BigscapeGCFLoader(bigscape_cluster_file)
            logger.info(f"Loading BigSCAPE cluster file {bigscape_cluster_file}")
        elif self.config.bigscape.version == 2:
            bigscape_db_file = self.config.root_dir / defaults.BIGSCAPE_DIRNAME / "data_sqlite.db"
            loader = BigscapeV2GCFLoader(bigscape_db_file)
            logger.info(f"Loading BigSCAPE database file {bigscape_db_file}")
        else:
            raise ValueError(f"Unsupported BigScape version: {self.config.bigscape.version}")

        raw_gcfs = loader.get_gcfs()

        # Step 5: add BGC objects to GCF
        all_gcfs_with_bgc, _, _ = add_bgc_to_gcf(all_bgcs_with_strain, raw_gcfs)

        # Step 6: get mibig bgcs and strains in use from GCFs
        mibig_strains_in_use = StrainCollection()
        if self.config.mibig.to_use:
            mibig_bgcs_in_use, mibig_strains_in_use = get_mibig_from_gcf(all_gcfs_with_bgc)
        else:
            mibig_bgcs_in_use = []

        # Step 7: set attributes with valid objects
        self.bgcs = antismash_bgcs_with_strain + mibig_bgcs_in_use
        self.gcfs = all_gcfs_with_bgc
        self.mibig_strains_in_use = mibig_strains_in_use

        logger.info("Loading genomics data completed\n")
        return True

    @deprecated(reason="To be refactored. It was used in the `self.load` method before.")
    def _load_class_info(self):
        """Load class match info (based on mibig) and chemical class predictions.

        Run CANOPUS if asked for. First sirius is run through docker, if this
        fails, it is run with a version present on the path.

        Return:
            True if everything completes
        """
        # load Class_matches with mibig info from data
        mibig_class_file = (
            NPLINKER_APP_DATA_DIR / "MIBiG2.0_compounds_with_AS_BGC_CF_NPC_classes.txt"
        )

        self.class_matches = ClassMatches(mibig_class_file)  # noqa

        # run canopus if canopus_dir does not exist
        should_run_canopus = self._config_docker.get("run_canopus", self.RUN_CANOPUS_DEFAULT)
        extra_canopus_parameters = self._config_docker.get(
            "extra_canopus_parameters", self.EXTRA_CANOPUS_PARAMS_DEFAULT
        )
        if should_run_canopus:
            # don't run canopus when canopus dir exists already
            if not os.path.isdir(self.canopus_dir):
                logger.info(
                    'Running CANOPUS! extra_canopus_parameters="{}"'.format(
                        extra_canopus_parameters
                    )
                )
                try:
                    run_canopus(self.gnps_mgf_file, self.canopus_dir, extra_canopus_parameters)  # noqa
                except Exception as e:
                    logger.warning(
                        'Failed to run CANOPUS on mgf file with docker, error was "{}"'.format(e)
                    )
                    logger.info("Trying to run CANOPUS again using SIRIUS from path")
                    try:
                        run_canopus(self.gnps_mgf_file, self.canopus_dir, extra_canopus_parameters)  # noqa
                    except Exception as e:
                        logger.warning(
                            'Again failed to run CANOPUS on mgf file using sirius from path, error was "{}"'.format(
                                e
                            )
                        )
            else:
                logger.info("Found CANOPUS dir, CANOPUS not run again!")

        # load Chem_class_predictions (canopus, molnetenhancer are loaded)
        chem_classes = ChemClassPredictions(self.canopus_dir, self.molnetenhancer_dir, self._root)  # noqa
        # if no mf classes transfer them from spectra (due to old style MN)
        if not chem_classes.canopus.mf_classes and chem_classes.canopus.spectra_classes:
            logger.info("Added chemical compound classes for MFs")
            chem_classes.canopus.transfer_spec_classes_to_mfs(self.mfs)
        # include them in loader
        self.chem_classes = chem_classes
        return True



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import multiprocessing
from multiprocessing import Process
import numpy
from numba import jit


@jit(nopython=True)
def k_vec(a_mat, b, gamma=0.01):
    # vectorised kernel function
    d_sq = numpy.sum(numpy.power((a_mat - b), 2), axis=1)
    return numpy.exp(-gamma * d_sq)


@jit(nopython=True)
def project_candidate_opt(index, fingerprint, latent, x_kernel_vector, latent_basis, gamma):
    fingerprint_kernel_vector = k_vec(latent, fingerprint.T, gamma)
    res = numpy.dot(numpy.dot(fingerprint_kernel_vector, latent_basis), x_kernel_vector)
    return res, fingerprint_kernel_vector


@jit(nopython=True)
def preprocess_candidate_fingerprint(fingerprint, latent, latent_basis, gamma):
    fingerprint_kernel_vector = k_vec(latent, fingerprint.T, gamma)
    res = numpy.dot(fingerprint_kernel_vector, latent_basis)
    return res


@jit(nopython=True)
def project_candidate_preprocessed(preprocessed_fingerprint, x_kernel_vector):
    return numpy.dot(preprocessed_fingerprint, x_kernel_vector)


def project_candidate_wrapper(
    index, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma
):
    return [
        project_candidate_opt(index, fingerprint, latent, x_kernel_vector, latent_basis, gamma)
        for fingerprint in candidate_fingerprints
    ]


@jit(nopython=True)
def project_candidate_kernel_opt(index, fingerprint_kernel_vector, x_kernel_vector, latent_basis):
    res = numpy.dot(numpy.dot(fingerprint_kernel_vector, latent_basis), x_kernel_vector)
    return res, fingerprint_kernel_vector


def project_candidate_kernel_wrapper(index, candidate_kernels, x_kernel_vector, latent_basis):
    return [
        project_candidate_kernel_opt(index, kernel_vector, x_kernel_vector, latent_basis)
        for kernel_vector in candidate_kernels
    ]


# @jit(nopython=True)
def rank_candidates_opt(
    index, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma
):
    candidate_distances, fingerprint_kernel_vectors = project_candidates_opt(
        index, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma
    )
    return numpy.argsort(numpy.array(candidate_distances))[::-1], fingerprint_kernel_vectors


def project_candidates_opt(
    index, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma
):
    candidate_projections = [
        project_candidate_opt(index, fingerprint, latent, x_kernel_vector, latent_basis, gamma)
        for fingerprint in candidate_fingerprints
    ]
    candidate_distances = [x[0] for x in candidate_projections]
    fingerprint_kernel_vectors = [x[1] for x in candidate_projections]
    return candidate_distances, fingerprint_kernel_vectors


def preprocess_candidates(candidate_fingerprints, latent, latent_basis, gamma):
    fingerprint_projections = [
        preprocess_candidate_fingerprint(fp, latent, latent_basis, gamma)
        for fp in candidate_fingerprints
    ]
    return fingerprint_projections


def project_candidates_preprocessed(preprocessed_candidates, x_kernel_vector):
    candidate_distances = [
        project_candidate_preprocessed(x, x_kernel_vector) for x in preprocessed_candidates
    ]
    return candidate_distances, []


def rank_candidate_kernel_opt(index, candidate_kernels, latent, x_kernel_vector, latent_basis):
    candidate_projections = [
        project_candidate_kernel_opt(index, kernel_vector, x_kernel_vector, latent_basis)
        for kernel_vector in candidate_kernels
    ]
    candidate_distances = [x[0] for x in candidate_projections]
    fingerprint_kernel_vectors = [x[1] for x in candidate_projections]
    return numpy.argsort(numpy.array(candidate_distances))[::-1], fingerprint_kernel_vectors


class InputOutputKernelRegression:
    def __init__(self, data):
        self.data = data
        self.kernel_vector_cache = {}

        self.ker_v_list = []

    def set_training_indices(self, indices=None, _lambda=0.1):
        self._lambda = _lambda
        if indices is None:
            indices = range(self.data.data_size)
        self.training_set = indices

    def fit(self):
        training_data_kernel = self.data.get_kernel_matrix(self.training_set)
        training_data_latent = self.data.get_latent_vectors(self.training_set)

        eye = numpy.eye(len(training_data_kernel))

        training_data_latent = numpy.array(training_data_latent).T

        latent_basis = numpy.linalg.inv(self._lambda * eye + training_data_kernel)
        self.latent_basis = latent_basis
        self.basis = numpy.dot(training_data_latent, latent_basis)

        self.latent = self.data.get_latent_vectors_vec(self.training_set)

    def calculate_fingerprint_kernel_vector(
        self, fingerprint, training_data_latent, kernel="gaussian"
    ):
        # if kernel == 'gaussian':
        #     def k_vec(a_mat, b, gamma=0.01):
        #         # vectorised kernel function
        #         d_sq = numpy.sum(numpy.power((a_mat - b), 2), axis=1)
        #         return numpy.exp(- gamma * d_sq)

        kernel_vector = k_vec(training_data_latent, fingerprint.T)
        return kernel_vector

    def project_candidate(self, index, fingerprint, latent, output=None):
        fingerprint_kernel_vector = self.calculate_fingerprint_kernel_vector(fingerprint, latent)
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)
        res = numpy.dot(numpy.dot(fingerprint_kernel_vector, self.latent_basis), x_kernel_vector)

        if output is not None:
            output.value = res
        return res

    def rank_candidates(self, index, candidate_fingerprints):
        latent = self.data.get_latent_vectors_vec(self.training_set)

        # candidate_distances = [self.project_candidate(index, fingerprint, latent) for fingerprint in candidate_fingerprints]

        candidate_distances = []
        batch_size = 100
        for i in range(int(len(candidate_fingerprints) / batch_size) + 1):
            if i % 10 == 0:
                print("processing set %s" % i)
            candidate_fingerprints_subset = candidate_fingerprints[
                i * batch_size : (i + 1) * batch_size
            ]
            values = [
                multiprocessing.Value(ctypes.c_float)
                for x in range(len(candidate_fingerprints_subset))
            ]
            candidate_distance_jobs = [
                Process(target=self.project_candidate, args=(index, fingerprint, latent, value))
                for fingerprint, value in zip(candidate_fingerprints_subset, values)
            ]
            [x.start() for x in candidate_distance_jobs]
            [x.join() for x in candidate_distance_jobs]
            candidate_distances_subset = [x.value for x in values]
            candidate_distances.extend(candidate_distances_subset)

        return [
            x[1]
            for x in sorted(
                zip(candidate_distances, range(len(candidate_distances))),
                key=lambda x: x[0],
                reverse=True,
            )
        ]

    def rank_candidates_opt(self, index, candidate_fingerprints, candidate_set_name=None):
        latent = self.latent
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)
        latent_basis = self.latent_basis
        gamma = 0.01

        candidate_ranking, fp_kernels = rank_candidates_opt(
            index, candidate_fingerprints, latent, x_kernel_vector, latent_basis, gamma
        )

        # if candidate_set_name is not None:
        #     with open('/home/grimur/iokr/data/candidate_hash/fp_gamma0.01_%s.bin' % candidate_set_name, 'wb') as f:
        #         pickle.dump(fp_kernels, f)

        return candidate_ranking

    def get_data_for_candidate_ranking(self, index):
        latent, latent_basis, gamma = self.get_data_for_novel_candidate_ranking()
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)

        return latent, x_kernel_vector, latent_basis, gamma

    def get_data_for_novel_candidate_ranking(self):
        latent = self.latent
        latent_basis = self.latent_basis
        gamma = 0.01
        return latent, latent_basis, gamma

    def project(self, index):
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)
        projection = numpy.dot(self.basis, x_kernel_vector)
        return projection

    def test(self, index, cutoff=0.01):
        proj = self.project(index)
        return [1 if x > cutoff else 0 for x in proj]

    def get_kernel_vector_for_sample(self, ms):
        kernel_vector = []

        ms_auto = self.data.calculate_kernel(ms, ms)
        for idx in self.training_set:
            t_sp = self.data.ms[idx]
            t_sp_auto = self.data.calculate_kernel(t_sp, t_sp)
            ms_t_sp = self.data.calculate_kernel(ms, t_sp)
            kernel_value = ms_t_sp / numpy.sqrt(ms_auto * t_sp_auto)
            kernel_vector.append(kernel_value)

        return kernel_vector


def main():
    target_vectors = numpy.array(
        [[0, 1, 0], [1, 0, 0], [1, 1, 0], [1, 1, 1], [0, 0, 1], [0, 1, 1], [1, 0, 1]], dtype="float"
    )
    repr_vectors = numpy.array(
        [[0, 1, 0], [1, 0, 0], [1, 1, 0], [1, 1, 1], [0, 0, 1], [0, 1, 1], [1, 0, 1]], dtype="float"
    )

    kernel_matrix = numpy.zeros((7, 7))
    for i in range(7):
        for j in range(i + 1):
            kernel_matrix[i, j] = kernel_matrix[j, i] = numpy.dot(repr_vectors[i], repr_vectors[j])

    data = DataStore(kernel_matrix, target_vectors)
    okr = InputOutputKernelRegression(data)
    okr.set_training_indices([0, 2, 3], _lambda=0.0)
    okr.fit()
    print(okr.test(0), target_vectors[0])
    print(okr.test(1), target_vectors[1])
    print(okr.test(2), target_vectors[2])
    print(okr.test(3), target_vectors[3])


if __name__ == "__main__":
    main()



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import random
import time
import iokr_opt
import iokrdata as iokrdataserver
import numpy
from iokrdata import load_kernels
from mk_fprints import fingerprint_from_smiles


def gather_results(active_jobs, limit=25):
    done_jobs = []
    while len(active_jobs) > limit:
        remaining_jobs = []
        for line in active_jobs:
            job = line[-1]
            if job.ready():
                res = job.get()
                new_line = list(line[:-1])
                new_line.append(res)
                done_jobs.append(new_line)
            else:
                remaining_jobs.append(line)
        active_jobs = remaining_jobs
        time.sleep(1)
    return active_jobs, done_jobs


def run_iokr(data):
    collected_rankings = []
    collected_baseline_rankings = []

    all_indices = data.get_all_indices()
    filtered_indices = all_indices
    # filtered_indices = numpy.load('/home/grimur/iokr/data/mibig/mibig_idx_to_skip.bin.npy')
    # filtered_indices = list(filtered_indices)

    iokr = iokr_opt.InputOutputKernelRegression(data)
    iokr.set_training_indices(filtered_indices, _lambda=0.001)
    iokr.fit()

    latent, latent_basis, gamma = iokr.get_data_for_novel_candidate_ranking()

    # init the common candidate set
    candidate_set = [x for x in data.candidate_set]
    # init the sample set (same as candidate set?)
    spectra_kernels = data.spectra_kernels

    for i, sample_kernel_vector in enumerate(spectra_kernels):
        sample_kernel_vector = sample_kernel_vector[filtered_indices]
        # match the sample to the candidate set
        correct_indices = data.spectra_to_bgc_indices[i]
        candidate_fingerprints = numpy.array([x[1] for x in candidate_set]).copy()
        [x[0] for x in candidate_set]

        args = (0, candidate_fingerprints, latent, sample_kernel_vector, latent_basis, gamma)
        with open("secondary.bin", "wb") as f:
            import pickle

            pickle.dump(args, f)
        output = iokr_opt.rank_candidates_opt(*args)

        res_i = i
        res_correct_indices = correct_indices
        res_output = output

        res_ranking = list(res_output[0])

        # actual rankings
        correct_ranking = min(res_ranking.index(x) for x in res_correct_indices)
        collected_rankings.append((res_i, correct_ranking, len(res_ranking)))

        # correct_random_rankings = []
        # for random_iteration in range(100):
        #     random.shuffle(res_ranking)
        #     correct_random_ranking = min([res_ranking.index(x) for x in res_correct_indices])
        #     correct_random_rankings.append(correct_random_ranking)
        # collected_baseline_rankings.append((res_i, numpy.mean(correct_random_rankings), len(res_ranking)))

        random.shuffle(res_ranking)
        correct_random_ranking = min(res_ranking.index(x) for x in res_correct_indices)
        collected_baseline_rankings.append((res_i, correct_random_ranking, len(res_ranking)))

        total = len(collected_rankings)
        print(
            "%.06f, %.06f, %s"
            % (
                float([x[1] for x in collected_rankings].count(0)) / total,
                float([x[1] for x in collected_baseline_rankings].count(0)) / total,
                total,
            )
        )

        # print(cr_b[res_i], cr_a[res_i], cr_a[res_i] == cr_b[res_i])

    print("")
    print("IOKR test run done!")
    print(f"#samples: {len(collected_rankings)}")
    print(
        "top-1 acc: {}".format(
            float([x[1] for x in collected_rankings].count(0)) / total,
        )
    )
    print(
        "top-1 base: {}".format(
            float([x[1] for x in collected_baseline_rankings].count(0)) / total,
        )
    )

    return collected_rankings, collected_baseline_rankings


def get_test_set(fingerprint):
    import csv
    from pyteomics import mgf

    mapping = []
    # TODO: fix the path
    with open("/home/grimur/iokr/data/mibig/matched_mibig_gnps_2.0.csv") as f:
        for line in csv.reader(f):
            if not line[0].startswith("#"):
                mapping.append(line)

    mapping_dict = {}
    bgc_id_list = []
    bgc_fps = {}
    for line in mapping:
        line_id = line[0] + line[1]
        if line[0] in mapping_dict:
            mapping_dict[line_id].append(line)
        else:
            mapping_dict[line_id] = [line]

        bgc_id = line[4]
        bgc_smiles = line[7]
        bgc_fp = fingerprint_from_smiles(bgc_smiles, fingerprint)
        if bgc_id not in bgc_id_list:
            bgc_id_list.append(bgc_id)
            bgc_fps[bgc_id] = bgc_fp

    bgc_fp_list = [bgc_fps[x] for x in bgc_id_list]
    candidate_set = zip(bgc_id_list, bgc_fp_list)

    mgf_file = "/home/grimur/iokr/data/mibig/matched_mibig_gnps_2.0.mgf"
    spectra_to_bgc_indices = []
    for i, spectrum in enumerate(mgf.read(mgf_file)):
        spec_file = spectrum["params"]["filename"]
        spec_id = spectrum["params"]["feature_id"]
        spec_filename = spec_file + spec_id
        bgc_matches = mapping_dict[spec_filename]
        bgc_indices = []
        for line in bgc_matches:
            bgc_id = line[4]
            bgc_idx = bgc_id_list.index(bgc_id)
            bgc_indices.append(bgc_idx)
        spectra_to_bgc_indices.append(bgc_indices)

    kernel_files = [
        #'/home/grimur/iokr/sml_test_test_nloss.npy',
        #'/home/grimur/iokr/sml_test_test_peaks.npy'
        "/home/grimur/iokr/mibig_test_nloss.npy",
        "/home/grimur/iokr/mibig_test_peaks.npy",
    ]
    spectra_kernels = load_kernels(kernel_files, normalise=False)

    return spectra_kernels, candidate_set, spectra_to_bgc_indices


def main():
    parser = argparse.ArgumentParser("Run IOKR test on a set")
    parser.add_argument("--kernel", dest="kernel", help="Kernel files", nargs="+")
    parser.add_argument(
        "--fp",
        dest="fingerprint",
        help="fingerprint type (substructure, cdk (default), klekota-roth",
        default="cdk_default",
    )
    parser.add_argument("--data", dest="datapath", help="data path", required=True)
    parser.add_argument("--output", dest="output", help="output label", required=True)
    args = parser.parse_args()

    # read from args
    # datapath = '/home/grimur/iokr/data'
    datapath = args.datapath
    # kernel_files = [datapath + os.sep + 'input_kernels_gh/ppk_dag_all_normalised_shifted_nloss.npy',
    #                 datapath + os.sep + 'input_kernels_gh/ppk_dag_all_normalised_shifted_peaks.npy']
    kernel_files = args.kernel
    # fingerprint = None

    fingerprint = args.fingerprint

    output_file = "IOKRranking_%s.bin" % args.output
    output_baseline = "IOKRranking_baseline_%s.bin" % args.output

    iokrdata = iokrdataserver.IOKRDataServer(datapath, kernel=None)
    kernel_matrix = load_kernels(kernel_files)
    iokrdata.kernel = kernel_matrix
    with open(datapath + os.sep + "ind_eval.txt") as f:
        raw_data = f.read()
        test_sample_indices = [int(x) - 1 for x in raw_data.strip().split()]
    iokrdata.test_sample_indices = test_sample_indices

    if fingerprint is not None:
        iokrdata.set_fingerprint(fingerprint)

    spectra_kernels, candidate_set, spectra_to_bgc_indices = get_test_set(fingerprint)
    with open("inputtest.bin", "wb") as f:
        import pickle

        pickle.dump((spectra_kernels, candidate_set, spectra_to_bgc_indices), f)
    # with open('inputtest.bin', 'rb') as f:
    #     import pickle
    #     spectra_kernels, candidate_set, spectra_to_bgc_indices = pickle.load(f)

    # print(spectra_kernels)
    # print(candidate_set)
    # print(spectra_to_bgc_indices)

    iokrdata.spectra_kernels = spectra_kernels
    iokrdata.candidate_set = candidate_set
    iokrdata.spectra_to_bgc_indices = spectra_to_bgc_indices

    print("run iokr")
    rankings, baseline = run_iokr(iokrdata)

    numpy.save(output_file, rankings)
    numpy.save(output_baseline, baseline)


if __name__ == "__main__":
    main()



from __future__ import annotations
from abc import ABC
from abc import abstractmethod
from os import PathLike
from .bgc import BGC
from .gcf import GCF


class BGCLoaderBase(ABC):
    """Abstract base class for BGC loader."""

    def __init__(self, data_dir: str | PathLike) -> None:
        """Initialize the BGC loader.

        Args:
            data_dir: Path to directory that contains BGC metadata files
                (.json) or full data genbank files (.gbk).
        """
        self.data_dir = str(data_dir)

    @abstractmethod
    def get_files(self) -> dict[str, str]:
        """Get path to BGC files.

        Returns:
            The key is BGC name and value is path to BGC file
        """

    @abstractmethod
    def get_bgcs(self) -> list[BGC]:
        """Get BGC objects.

        Returns:
            A list of BGC objects
        """


class GCFLoaderBase(ABC):
    """Abstract base class for GCF loader."""

    @abstractmethod
    def get_gcfs(self, keep_mibig_only: bool, keep_singleton: bool) -> list[GCF]:
        """Get GCF objects.

        Args:
            keep_mibig_only: True to keep GCFs that contain only MIBiG
                BGCs.
            keep_singleton: True to keep singleton GCFs. A singleton GCF
                is a GCF that contains only one BGC.

        Returns:
            A list of GCF objects
        """



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import json
import logging
import os
import re


logger = logging.getLogger(__name__)

# this will match strings like '...cluster001.gbk' or '...region022.gbk',
# and allow the number to be extracted easily
CLUSTER_REGION_REGEX = re.compile("(.+?)\\.(cluster|region)(\\d+).gbk$")

# This module contains a couple of parsers for antiSMASH knownclusterblast output
# files: one for the newer per-directory .JSON blob and one for the legacy
# format with one text file per gbk.


class KCBJSONParser:
    """Parser for the large .json files antiSMASH generates as part of its output.

    This is supposed to do the same job as the KCBTextParser class without relying
    on parsing the legacy-format text files.

    Args:
        bgcs: a list of NPLinker BGC objects created from the same antiSMASH output folder
    """

    def __init__(self, bgcs):
        # check all the linked files exist
        for bgc in bgcs:
            if not os.path.exists(bgc.antismash_file):
                raise Exception('KCBJSONParser failed to find file "{}"'.format(bgc.antismash_file))

        logger.info(f"KCBJSONParser({len(bgcs)} BGCs)")

        # find the JSON file: TODO is the assumption of there only being a single .json
        # file always going to work? otherwise have to try guessing the name based on
        # genome IDs
        prefix = os.path.dirname(bgcs[0].antismash_file)
        json_files = list(filter(lambda f: f.endswith(".json"), os.listdir(prefix)))
        logger.info("Found {} JSON files in {}".format(len(json_files), prefix))

        if len(json_files) == 0:
            logger.warning("Unable to find an antiSMASH JSON output file in {}".format(prefix))
            self.json_filename = None
            return

        self.json_filename = os.path.join(prefix, json_files[0])
        logger.info(f"Using JSON file {self.json_filename}")

    def parse_hits(self):
        if self.json_filename is None:
            return None

        self.collected_hits = {}

        # NOTE: since some of these JSON files are ridiculously large (>50MB in some cases)
        # it can cause OOM errors (with the stdlib json module and a couple of other
        # 3rd party ones). Best way to avoid this is by using 64-bit Python.
        # TODO: catch exception and print message recommending 64-bit
        logger.info("Loading antiSMASH JSON data from {}".format(self.json_filename))
        data = json.load(open(self.json_filename))

        # JSON data structure: depending on the gbks in the source folder, the
        # structure of the results may be different from one to the next. from
        # looking at some examples, what should be expected is a structure like
        # this outline ([] = array/list object):
        #   <top>
        #       [records]
        #           id
        #           name
        #           description
        #           [features]
        #           [modules]
        #               ...
        #               [antismash.modules.clusterblast] (may be missing)
        #                   record_id
        #                   [knowncluster]
        #                       record_id
        #                       [results]
        #                           region_number
        #                           total_hits
        #                           [ranking] (may be empty if total_hits = 0)
        #                               (tuple_pt1, tuple_pt2)
        #                               ...
        #                       mibig_entries
        #                           (lists of IDs named by region numbers)
        #
        # every gbk with a distinct prefix maps to an entry in the [records] list.
        # however there are also entries for gbks which don't exist in the source
        # folder, don't know why. These can easily be excluded by checking for
        # the presence of the "antismash.modules.clusterblast" module results in
        # the [modules] list for that record.
        #
        # the next step is to access the [results] list nested further down in
        # the structure. the number of elements in this list are dependent on the
        # number of gbks that are different regions of the same genome. e.g. if
        # you have foobar.region001.gbk and foobar.region002.gbk, there should be
        # a single 'record' for 'foobar' and a 'result' for each region. The
        # region_number field can easily be used to distinguish these.
        #
        # finally there is the 'ranking' list. this is where all the data we need
        # is located. it has a weird structure where every entry is a 2-tuple with
        # this content:
        #   1st element: has fields accession, cluster_label, description, and
        #       cluster_type, plus lists 'proteins' and 'tags'
        #   2nd element: has fields hits, blast_score, core_gene_hits,
        #       synteny_score, core_bonus, and a 'pairings' list
        #
        # each element of the 'pairings' list is in turn a 3-tuple:
        #   1st element: a |-delimited field with multiple values in it. a typical
        #       example is 'input|c1|0-2665|-|B098_RS0100005|indolepyruvate'
        #   2nd element: a single integer, which seems to be the 0-based index
        #       of this 'pairing' in the list?
        #   3rd element: an object with fields name, genecluster, start, end,
        #       strand, annotation, perc_ident, blastscore, perc_coverage, evalue,
        #       and locus_tag
        #
        #  to correctly parse this, need to iterate over the list of records, and
        #  for each one that has clusterblast results iterate over those, matching
        #  things up to the BGC objects created by nplinker as we go along.

        for i, rec in enumerate(data["records"]):
            hits = self._parse(rec)
            if hits is not None:
                self.collected_hits.update(hits)

        logger.info(
            "KCBJSONParser: collected {} total hit entries".format(len(self.collected_hits))
        )

        return self.collected_hits

    def _parse(self, record):
        """Parses the knownclusterblast data for a single 'record' entry."""
        modules = record.get("modules", None)
        if modules is None or "antismash.modules.clusterblast" not in modules:
            # this probably isn't an error, the JSON often seems to contain entries
            # for gbks that don't exist in the source folder (??)
            return None

        kcb = modules["antismash.modules.clusterblast"]["knowncluster"]
        record_id = kcb["record_id"]

        # each 'record' may contain multiple results, and each 'result' may be
        # linked to a different BGC and region number, so need to keep track of
        # all this to match them to NPLinker BGC objects later
        record_hits = {}

        for i, result in enumerate(kcb["results"]):
            # check number of rankings, can ignore if zero
            if len(result["ranking"]) == 0:
                continue

            # step 1: extract the keys of the 'mibig_entries' dict for the current region number (there should be
            # the same number of 'results' as 'mibig_entries'). There appear to be fewer entries in these lists than
            # appear in the text files where both exist, but I can't find another source for them and the extras
            # don't appear to matter much as they aren't referenced later in the text files or by the original parser
            gene_ids = list(kcb["mibig_entries"][str(i + 1)].keys())

            # step 2: extract the MiBIG BGC IDs + product names (=parsing the Significant hits table in text file)
            sig_hits = [
                (entry[0]["accession"], entry[0]["description"]) for entry in result["ranking"]
            ]

            # give up if there aren't any of these
            if len(sig_hits) == 0:
                continue

            # step 3: extract BGC IDs (=parsing of "... subject cluster" tables in text file)
            # (this is a list of lists)
            all_mibig_genes = [entry[0]["proteins"] for entry in result["ranking"]]

            region_number = int(result["region_number"])
            # should never have two identical region numbers for the same ID
            assert region_number not in record_hits

            # want to construct the same data structure as the original parser
            # should contain:
            #   - all_bgc_genes => gene_ids
            #   - all_mibig_genes => all_mibig_genes
            #   - invididual_hits => list populated based on blast_hits
            hit = {}
            assert len(all_mibig_genes) == len(result["ranking"])

            # step 4: extract the blast hit information from the 'rankings' and
            # 'pairings' lists (=parsing of "Table of Blast hits" in text file)

            # each "hit" shares these
            hit = {"all_bgc_genes": gene_ids}
            individual_hits = []
            for j, ranking in enumerate(result["ranking"]):
                hit["all_mibig_genes"] = all_mibig_genes[j]
                hit["mibig_id"] = sig_hits[j][0]
                for pairing in ranking[1]["pairings"]:
                    tc1 = pairing[0].split("|")[4]
                    individual_hits.append(
                        {
                            "source_bgc_gene": tc1,
                            "mibig_bgc_gene": pairing[2]["name"],
                            "identity_percent": int(pairing[2]["perc_ident"]),
                            "blast_score": int(pairing[2]["blastscore"]),
                        }
                    )

            hit["individual_hits"] = individual_hits
            record_hits[region_number] = hit
            # logger.debug('\tSig hit {}/{} for {}: found {} individual hits'.format(j+1, len(sig_hits), sig_hits[j][0], len(individual_hits)))

        # don't create entries with no results
        if len(record_hits) == 0:
            return {}

        return {record_id: record_hits}


class KCBTextParser:
    """Parser for antismash knownclusterblast text output files."""

    def __init__(self, filename):
        if not os.path.exists(filename):
            raise Exception(f'KCBTextParser failed to find file "{filename}"')

        self.bgc_genes = set()
        self.mibig_bgcs = []
        self.hits = {}

        # Structure outline for quick reference:
        # <top of file>
        # Clusterblast scores for <ID>
        #
        # Table of genes, locations, strands and annotations of query cluster:
        # <multiple lines in table> (this is top_block below)
        #
        # Significant hits:
        # <multiple hits> (this is second_block below)
        #
        # Details:
        #
        # >>
        # <multiple lines of details for hit 1>
        #
        # >>
        # <multiple lines of details for hit 2>
        #
        # (continues up to hit n)
        # EOF

        with open(filename) as f:
            line = next(f)
            while not line.startswith("Table of genes"):
                line = next(f)
            # now we're in the first block. this is the section beginning with the
            # line "Table of genes, locations, strands and annotations of query cluster:"
            # and ending at "Significant hits:"
            top_block = []
            while True:
                line = next(f)
                if line.startswith("Significant"):
                    break
                else:
                    if len(line) > 1:
                        top_block.append(line.rstrip())

            # now we're in the second block. this is the section beginning with "Significant hits:"
            # and ending at "Details:"
            second_block = []
            while True:
                line = next(f)
                if line.startswith("Details"):
                    break
                else:
                    if len(line) > 1:
                        second_block.append(line.rstrip())

            # find the start of the first ">>" delimited section
            while True:
                try:
                    line = next(f)
                    if line.startswith(">>"):
                        break
                except Exception:
                    # EOF
                    return

            details = []
            finished = False
            while not finished:
                temp_list = []
                while True:
                    try:
                        line = next(f)
                        if line.startswith(">>"):
                            # finished current section
                            details.append(temp_list)
                            break
                        else:
                            if len(line) > 1:
                                temp_list.append(line.rstrip())
                    except Exception:
                        details.append(temp_list)
                        finished = True
                        break

            # do some processing on the blocks
            # firstly, extract the genes from the BGC -- stored in the first block
            # these are the gene names in the BGC
            # e.g. given a line "STROP_RS12480	2788559	2789786	-	cytochrome P450	"
            # this would add "STROP_RS12480" to self.bgc_genes
            for line in top_block:
                tokens = line.split()
                self.bgc_genes.add(tokens[0])

            # secondly, extract the MiBIG BGCs that are mentioned here
            # e.g. given a line "1. BGC0000279_c1	Xantholipin_biosynthetic_gene_cluster"
            # this would take "BGC0000279_c1" as the bgc_id and "Xantholipin_biosynthetic_gene_cluster"
            # as the bgc_product_name
            for line in second_block:
                tokens = line.split()
                bgc_id = tokens[1]
                bgc_product_name = tokens[2]
                self.mibig_bgcs.append((bgc_id, bgc_product_name))

            # details is a list of lists, where each sublist contains the set of lines
            # for one ">>" delimited section giving the detailed information on a
            # particular hit
            for i, detail in enumerate(details):
                # the first line is of the form "1. BGC0000279_c1"
                current_bgc_id = detail[0].split()[1]  # this is the MiBIG ID, e.g. BGC0001666

                self.hits[current_bgc_id] = {}
                self.hits[current_bgc_id]["all_bgc_genes"] = self.bgc_genes

                # if the detail sections don't appear in the same order as the "Significant hits"
                # section then something is wrong and can't continue parsing this
                if current_bgc_id != self.mibig_bgcs[i][0]:
                    raise Exception(
                        "Mismatched ordering found in knownclusterblast file {}".format(filename)
                    )

                # get the line numbers where these two tables start
                table_pos = detail.index(
                    "Table of genes, locations, strands and annotations of subject cluster:"
                )
                pos = detail.index(
                    "Table of Blast hits (query gene, subject gene, %identity, blast score, %coverage, e-value):"
                )

                # get the list of all of the genes within the MiBIG BGC by iterating over
                # the lines in the "Table of genes ..." table
                all_mibig_genes = []
                for line in detail[table_pos + 1 : pos]:
                    all_mibig_genes.append(line.split()[0])

                if len(all_mibig_genes) == 0:
                    logger.warning(
                        "KCBTextParser failed to extract any MiBIG genes from file {}, BGC ID {}".format(
                            filename, current_bgc_id
                        )
                    )
                    # just continue to the next section here because this currently makes the rest of the processing invalid
                    del self.hits[current_bgc_id]
                    continue

                self.hits[current_bgc_id]["all_mibig_genes"] = all_mibig_genes
                self.hits[current_bgc_id]["individual_hits"] = []

                for line in detail[pos + 1 :]:
                    tokens = line.split()
                    bgc_id = tokens[0]
                    self.hits[current_bgc_id]["individual_hits"].append(
                        {
                            "source_bgc_gene": tokens[0],
                            "mibig_bgc_gene": tokens[1],
                            "identity_percent": int(tokens[2]),
                            "blast_score": int(tokens[3]),
                        }
                    )

    @staticmethod
    def get_kcb_filename_from_bgc(bgc):
        """Given a BGC object, return the filename of the corresponding knownclusterblast .txt file (if any).

        This method attempts to derive the name of the knownclusterblast output file for the
        supplied BGC object, using the original path + filename of the .gbk that the BGC was
        sourced from during the loading process.

        Assumptions made:
            - the .gbk files are grouped in subdirectories with <dataset>/antismash/
            - each of these subdirectories contains a "knownclusterblast" subdirectory
            - within that, there is a single .txt file for each .gbk

        For some reason there doesn't appear to be a simple way of matching up the .gbk/.txt
        files using their contents, only the filenames. And there are various different
        possibilities for the naming schemes, so this method tries to account for the
        known variants discovered so far.

        Args:
            bgc: a BGC object

        Returns:
            A string containing the filename of the knownclusterblast .txt file for the
            supplied BGC, or None if an error occurred/file doesn't exist
        """
        if bgc.antismash_file is None:
            logger.warning(f"BGC {bgc} has no antismash_file set")
            return None

        # expecting to find the .txt files inside a 'knownclusterblast' subdir in the
        # same location as the .gbk file, give up if that doesn't exist
        base_path = os.path.join(os.path.dirname(bgc.antismash_file), "knownclusterblast")

        if not os.path.exists(base_path):
            logger.warning(
                'Expected "knownclusterblast" directory not found at "{}"'.format(base_path)
            )
            return None

        # get the name of the .gbk file itself (no path)
        genbank_file = os.path.split(bgc.antismash_file)[1]

        if "region" in genbank_file:
            # case 1: assume the genbank files have a <someID>.region<num>.gbk naming scheme.
            # this seems to be the case for all the more recent datasets i've seen.

            # use a regex to extract the portions of the filename we're interested in
            regex_obj = CLUSTER_REGION_REGEX.search(genbank_file)
            number = int(regex_obj.group(3))
            prefix = regex_obj.group(1)

            # construct the expected filename using the prefix and number
            # (note no leading zeroes on the number)
            kcb_name = os.path.join(base_path, f"{prefix}_c{number}.txt")
        elif "cluster" in genbank_file:
            # case 2: assume the genbank files have a <someID>.cluster<num>.gbk naming scheme.
            # this is the case with the Crusemann dataset on the paired platform among others.

            # use a regex to extract the number from the filename
            regex_obj = CLUSTER_REGION_REGEX.search(genbank_file)
            number = int(regex_obj.group(3))

            # construct the expected filename
            # (note no leading zeroes on the number)
            kcb_name = os.path.join(base_path, f"cluster{number}.txt")
        else:
            logger.warning(
                "Unknown GenBank file naming scheme, failed to determine knownclusterblast filenames!"
            )
            return None

        return kcb_name



from __future__ import annotations
import csv
from os import PathLike
from pathlib import Path
from nplinker.metabolomics.abc import FileMappingLoaderBase
from nplinker.utils import is_file_format
from .gnps_format import GNPSFormat
from .gnps_format import gnps_format_from_file_mapping


class GNPSFileMappingLoader(FileMappingLoaderBase):
    """Class to load file mappings from GNPS output file.

    ??? info "Concept"
        [GNPS data][gnps-data]

    File mappings refers to the mapping from spectrum id to files in which
    this spectrum occurs.

    The file mappings file is from GNPS output archive, as described below
    for each GNPS workflow type:

    1. METABOLOMICS-SNETS
        - clusterinfosummarygroup_attributes_withIDs_withcomponentID/*.tsv
    2. METABOLOMICS-SNETS-V2
        - clusterinfosummarygroup_attributes_withIDs_withcomponentID/*.clustersummary
    3. FEATURE-BASED-MOLECULAR-NETWORKING
        - quantification_table*/*.csv
    """

    def __init__(self, file: str | PathLike) -> None:
        """Initialize the GNPSFileMappingLoader.

        Args:
            file: Path to the GNPS file mappings file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.

        Examples:
            >>> loader = GNPSFileMappingLoader("gnps_file_mappings.tsv")
            >>> print(loader.mappings["1"])
            ['26c.mzXML']
            >>> print(loader.mapping_reversed["26c.mzXML"])
            {'1', '3', '7', ...}
        """
        self._gnps_format = gnps_format_from_file_mapping(file)
        if self._gnps_format is GNPSFormat.Unknown:
            raise ValueError("Unknown workflow type for GNPS file mappings file ")

        self._file = Path(file)
        self._mapping: dict[str, list[str]] = {}

        self._validate()
        self._load()

    @property
    def mappings(self) -> dict[str, list[str]]:
        """Return mapping from spectrum id to files in which this spectrum occurs.

        Returns:
            Mapping from spectrum id to names of all files in which this spectrum occurs.
        """
        return self._mapping

    @property
    def mapping_reversed(self) -> dict[str, set[str]]:
        """Return mapping from file name to all spectra that occur in this file.

        Returns:
            Mapping from file name to all spectra ids that occur in this file.
        """
        mapping_reversed: dict[str, set[str]] = {}
        for spectrum_id, ms_filenames in self._mapping.items():
            for filename in ms_filenames:
                if filename in mapping_reversed:
                    mapping_reversed[filename].add(spectrum_id)
                else:
                    mapping_reversed[filename] = {spectrum_id}

        return mapping_reversed

    def _validate(self) -> None:
        """Validate the file mappings file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.
        """
        # validate file format
        required_file_formats = {
            GNPSFormat.SNETS: "tsv",
            GNPSFormat.SNETSV2: "tsv",
            GNPSFormat.FBMN: "csv",
        }
        if not is_file_format(self._file, required_file_formats[self._gnps_format]):
            raise ValueError(
                f"Invalid GNPS file mappings file '{self._file}'. "
                f"Expected a {required_file_formats[self._gnps_format]} file."
            )

        # validate required columns against the header
        required_columns = {
            GNPSFormat.SNETS: ["cluster index", "AllFiles"],
            GNPSFormat.SNETSV2: ["cluster index", "UniqueFileSources"],
            GNPSFormat.FBMN: ["row ID", " Peak area"],
        }
        with open(self._file, mode="rt") as f:
            header = f.readline()
            for k in required_columns[self._gnps_format]:
                if k not in header:
                    raise ValueError(
                        f"Invalid GNPS file mappings file '{self._file}'. "
                        f"Expected a header line with '{k}' column, "
                        f"but got '{header}'."
                    )

        # validate that cluster index or row id must be unique
        with open(self._file, mode="rt") as f:
            if self._gnps_format is GNPSFormat.FBMN:
                reader = csv.DictReader(f, delimiter=",")
                ids = [row["row ID"] for row in reader]
            else:
                reader = csv.DictReader(f, delimiter="\t")
                ids = [row["cluster index"] for row in reader]
        duplicates = {x for x in ids if ids.count(x) > 1}
        if len(duplicates) > 0:
            raise ValueError(
                f"Invalid GNPS file mappings file '{self._file}'. "
                f"Expected unique 'cluster index' or 'row ID', "
                f"but found duplicates '{duplicates}'."
            )

    def _load(self) -> None:
        """Load file mapping from the file based on the GNPS workflow type."""
        if self._gnps_format is GNPSFormat.SNETS:
            self._load_snets()
        elif self._gnps_format is GNPSFormat.SNETSV2:
            self._load_snetsv2()
        elif self._gnps_format is GNPSFormat.FBMN:
            self._load_fbmn()

    def _load_snets(self) -> None:
        """Load file mapping from output of GNPS SNETS workflow.

        The following columns are loaded:

        - "cluster index": loaded as spectrum id
        - "AllFiles": a list of files in which the spectrum occurs, separated
            by '###'.
            An example data of "AllFiles" column is as follows:
            "2b.mzXML:1503195###6a.mzXML:1502983###"
        """
        with open(self._file, mode="rt", encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                spectrum_id = row["cluster index"]
                occurrences = row["AllFiles"].split("###")  # split by '###'
                occurrences.pop()  # remove last empty entry
                # separate the scan position from the files
                samples = [x.split(":")[0] for x in occurrences]
                self._mapping[spectrum_id] = samples

    def _load_snetsv2(self) -> None:
        """Load file mapping from output of GNPS SNETS-V2 workflow.

        The following columns are loaded:

        - "cluster index": loaded as spectrum id
        - "UniqueFileSources": a list of files in which the spectrum occurs,
            separated by '|'.
            An example data of "UniqueFileSources" column is as follows:
            "140221_Blanc5.mzML|140221_Blanc8.mzML|140221_ME_14_12.mzML"
        """
        with open(self._file, mode="rt", encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                spectrum_id = row["cluster index"]
                samples = row["UniqueFileSources"].split("|")
                self._mapping[spectrum_id] = samples

    def _load_fbmn(self) -> None:
        """Load file mapping from output of GNPS FBMN workflow.

        The column "row ID" is loaded as spectrum id.

        The column names containing " Peak area" are used to extract the file
        names, and the values of these columns are used to determine whether
        the spectrum occurs in the file. The file name is taken only if the
        value is greater than 0.

        An example data of the file is as follows:
            ```
            row ID,5434_5433_mod.mzXML Peak area,5425_5426_mod.mzXML Peak area
            1,1764067.8434999974,0.0
            ```
        """
        pattern = " Peak area"
        with open(self._file, mode="rt", encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter=",")
            for row in reader:
                spectrum_id = row["row ID"]
                samples = []
                for col in row:
                    if pattern in col and float(row[col]) > 0:
                        samples.append(col.strip(pattern))
                self._mapping[spectrum_id] = samples



from __future__ import annotations
import fnmatch
import json
import shutil
from glob import glob
from os import PathLike
from pathlib import Path
from dynaconf import Dynaconf
from jsonschema import validate
import nplinker.defaults as defaults
from nplinker.genomics.antismash import podp_download_and_extract_antismash_data
from nplinker.genomics.bigscape.runbigscape import run_bigscape
from nplinker.genomics.mibig import download_and_extract_mibig_metadata
from nplinker.genomics.utils import generate_mappings_genome_id_bgc_id
from nplinker.metabolomics.gnps import GNPSDownloader
from nplinker.metabolomics.gnps import GNPSExtractor
from nplinker.schemas import STRAIN_MAPPINGS_SCHEMA
from nplinker.schemas import USER_STRAINS_SCHEMA
from nplinker.schemas import validate_podp_json
from nplinker.strain.utils import podp_generate_strain_mappings
from nplinker.utils import download_url
from nplinker.utils import list_dirs
from nplinker.utils import list_files


PODP_PROJECT_URL = "https://pairedomicsdata.bioinformatics.nl/api/projects/{}"


class DatasetArranger:
    """Arrange datasets based on the fixed working directory structure with the given configuration.

    ??? info "Concept and Diagram"
        [Working Directory Structure][working-directory-structure]

        [Dataset Arranging Pipeline][dataset-arranging-pipeline]

    "Arrange datasets" means:

    - For `local` mode (`config.mode` is `local`), the datasets provided by users are validated.
    - For `podp` mode (`config.mode` is `podp`), the datasets are automatically downloaded or
        generated, then validated.

    The datasets include MIBiG, GNPS, antiSMASH, and BiG-SCAPE data.

    Attributes:
        config: A Dynaconf object that contains the configuration settings.
        root_dir: The root directory of the datasets.
        downloads_dir: The directory to store downloaded files.
        mibig_dir: The directory to store MIBiG metadata.
        gnps_dir: The directory to store GNPS data.
        antismash_dir: The directory to store antiSMASH data.
        bigscape_dir: The directory to store BiG-SCAPE data.
        bigscape_running_output_dir: The directory to store the running output of BiG-SCAPE.
    """

    def __init__(self, config: Dynaconf) -> None:
        """Initialize the DatasetArranger.

        Args:
            config: A Dynaconf object that contains the configuration settings.


        Examples:
            >>> from nplinker.config import load_config
            >>> from nplinker.arranger import DatasetArranger
            >>> config = load_config("nplinker.toml")
            >>> arranger = DatasetArranger(config)
            >>> arranger.arrange()

        See Also:
            [DatasetLoader][nplinker.loader.DatasetLoader]: Load all data from files to memory.
        """
        self.config = config
        self.root_dir = config.root_dir
        self.downloads_dir = self.root_dir / defaults.DOWNLOADS_DIRNAME
        self.downloads_dir.mkdir(exist_ok=True)

        self.mibig_dir = self.root_dir / defaults.MIBIG_DIRNAME
        self.gnps_dir = self.root_dir / defaults.GNPS_DIRNAME
        self.antismash_dir = self.root_dir / defaults.ANTISMASH_DIRNAME
        self.bigscape_dir = self.root_dir / defaults.BIGSCAPE_DIRNAME
        self.bigscape_running_output_dir = (
            self.bigscape_dir / defaults.BIGSCAPE_RUNNING_OUTPUT_DIRNAME
        )

        self.arrange_podp_project_json()

    def arrange(self) -> None:
        """Arrange all datasets according to the configuration.

        The datasets include MIBiG, GNPS, antiSMASH, and BiG-SCAPE.
        """
        # The order of arranging the datasets matters, as some datasets depend on others
        self.arrange_mibig()
        self.arrange_gnps()
        self.arrange_antismash()
        self.arrange_bigscape()
        self.arrange_strain_mappings()
        self.arrange_strains_selected()

    def arrange_podp_project_json(self) -> None:
        """Arrange the PODP project JSON file.

        This method only works for the `podp` mode. If the JSON file does not exist, download it
        first; then the downloaded or existing JSON file will be validated according to the
        [PODP_ADAPTED_SCHEMA][nplinker.schemas.PODP_ADAPTED_SCHEMA].
        """
        if self.config.mode == "podp":
            file_name = f"paired_datarecord_{self.config.podp_id}.json"
            podp_file = self.downloads_dir / file_name
            if not podp_file.exists():
                download_url(
                    PODP_PROJECT_URL.format(self.config.podp_id),
                    self.downloads_dir,
                    file_name,
                )

            with open(podp_file, "r") as f:
                json_data = json.load(f)
            validate_podp_json(json_data)

    def arrange_mibig(self) -> None:
        """Arrange the MIBiG metadata.

        If `config.mibig.to_use` is `True`, download and extract the MIBiG metadata and override
        the existing MIBiG metadata if it exists. This ensures that the MIBiG metadata is always
        up-to-date to the specified version in the configuration.
        """
        if self.config.mibig.to_use:
            if self.mibig_dir.exists():
                # remove existing mibig data
                shutil.rmtree(self.mibig_dir)
            download_and_extract_mibig_metadata(
                self.downloads_dir,
                self.mibig_dir,
                version=self.config.mibig.version,
            )

    def arrange_gnps(self) -> None:
        """Arrange the GNPS data.

        For `local` mode, validate the GNPS data directory.

        For `podp` mode, if the GNPS data does not exist, download it; if it exists but not valid,
        remove the data and re-downloads it.

        The validation process includes:

        - Check if the GNPS data directory exists.
        - Check if the required files exist in the GNPS data directory, including:
            - `file_mappings.tsv` or `file_mappings.csv`
            - `spectra.mgf`
            - `molecular_families.tsv`
            - `annotations.tsv`
        """
        pass_validation = False
        if self.config.mode == "podp":
            # retry downloading at most 3 times if downloaded data has problems
            for _ in range(3):
                try:
                    validate_gnps(self.gnps_dir)
                    pass_validation = True
                    break
                except (FileNotFoundError, ValueError):
                    # Don't need to remove downloaded archive, as it'll be overwritten
                    shutil.rmtree(self.gnps_dir, ignore_errors=True)
                    self._download_and_extract_gnps()

        if not pass_validation:
            validate_gnps(self.gnps_dir)

        # get the path to file_mappings file (csv or tsv)
        self.gnps_file_mappings_file = self._get_gnps_file_mappings_file()

    def _get_gnps_file_mappings_file(self) -> Path:
        """Get the GNPS file mappings file.

        The GNPS file mappings file can be either a TSV file or a CSV file. This method checks if
        the TSV file or the CSV file exists in the default GNPS directory.

        Returns:
            Path to the GNPS file mappings file.
        """
        file_mappings_tsv = self.gnps_dir / defaults.GNPS_FILE_MAPPINGS_TSV
        file_mappings_csv = self.gnps_dir / defaults.GNPS_FILE_MAPPINGS_CSV

        gnps_file_mappings_file = (
            file_mappings_tsv if file_mappings_tsv.exists() else file_mappings_csv
        )

        return gnps_file_mappings_file  # type: ignore

    def _download_and_extract_gnps(self) -> None:
        """Download and extract the GNPS data.

        Get the GNPS task ID from the PODP project JSON file, then download and extract the GNPS
        data to the default GNPS directory.
        """
        podp_file = self.downloads_dir / f"paired_datarecord_{self.config.podp_id}.json"
        with open(podp_file, "r") as f:
            podp_json_data = json.load(f)
        gnps_task_id = podp_json_data["metabolomics"]["project"].get("molecular_network")

        data_archive = (
            GNPSDownloader(gnps_task_id, self.downloads_dir).download().get_download_file()
        )
        GNPSExtractor(data_archive, self.gnps_dir)

    def arrange_antismash(self) -> None:
        """Arrange the antiSMASH data.

        For `local` mode, validate the antiSMASH data.

        For `podp` mode, if the antiSMASH data does not exist, download it; if it exists but not
        valid, remove the data and re-download it.

        The validation process includes:

        - Check if the antiSMASH data directory exists.
        - Check if the antiSMASH data directory contains at least one sub-directory, and each
            sub-directory contains at least one BGC file (with the suffix `.region???.gbk` where
            `???` is a number).

        AntiSMASH BGC directory must follow the structure below:
        ```
        antismash
            ├── genome_id_1 (one AntiSMASH output, e.g. GCF_000514775.1)
            │  ├── GCF_000514775.1.gbk
            │  ├── NZ_AZWO01000004.region001.gbk
            │  └── ...
            ├── genome_id_2
            │  ├── ...
            └── ...
        ```
        """
        pass_validation = False
        if self.config.mode == "podp":
            for _ in range(3):
                try:
                    validate_antismash(self.antismash_dir)
                    pass_validation = True
                    break
                except FileNotFoundError:
                    shutil.rmtree(self.antismash_dir, ignore_errors=True)
                    self._download_and_extract_antismash()

        if not pass_validation:
            validate_antismash(self.antismash_dir)

    def _download_and_extract_antismash(self) -> None:
        """Download and extract the antiSMASH data.

        Get the antiSMASH data from the PODP project JSON file, then download and extract the
        antiSMASH data to the default antiSMASH directory.
        """
        podp_file = self.downloads_dir / f"paired_datarecord_{self.config.podp_id}.json"
        with open(podp_file, "r") as f:
            podp_json_data = json.load(f)
        podp_download_and_extract_antismash_data(
            podp_json_data["genomes"], self.downloads_dir, self.root_dir
        )

    def arrange_bigscape(self) -> None:
        """Arrange the BiG-SCAPE data.

        For `local` mode, if the BiG-SCAPE data is provided by users, validate it and raise an error
        if it is invalid. If the BiG-SCAPE data does not exist, run BiG-SCAPE to generate the
        clustering file.

        For `podp` mode, if the BiG-SCAPE data does not exist, run BiG-SCAPE to generate the
        clustering file; if it exists but not valid, remove the data and re-run BiG-SCAPE to generate
        the data.

        The running output of BiG-SCAPE will be saved to the directory `bigscape_running_output`
        in the default BiG-SCAPE directory, and the clustering file
        `mix_clustering_c{self.config.bigscape.cutoff}.tsv` will be copied to the default BiG-SCAPE
        directory.

        The validation process includes:

        - Check if the default BiG-SCAPE data directory exists.
        - Check if the clustering file `mix_clustering_c{self.config.bigscape.cutoff}.tsv` exists in the
                BiG-SCAPE data directory.
        - Check if the `data_sqlite.db` file exists in the BiG-SCAPE data directory.
        """
        if self.config.mode == "local" and self.bigscape_dir.exists():
            validate_bigscape(self.bigscape_dir, self.config.bigscape.cutoff)
            return

        pass_validation = False
        for _ in range(3):
            try:
                validate_bigscape(self.bigscape_dir, self.config.bigscape.cutoff)
                pass_validation = True
                break
            except FileNotFoundError:
                shutil.rmtree(self.bigscape_dir, ignore_errors=True)
                self._run_bigscape()

        if not pass_validation:
            validate_bigscape(self.bigscape_dir, self.config.bigscape.cutoff)

    def _run_bigscape(self) -> None:
        """Run BiG-SCAPE to generate the clustering file.

        The running output of BiG-SCAPE will be saved to the `BIGSCAPE_RUNNING_OUTPUT_PATH`.

        The clustering file "mix_clustering_c{self.config.bigscape.cutoff}.tsv" will be copied to the
        default BiG-SCAPE directory.
        """
        self.bigscape_running_output_dir.mkdir(exist_ok=True, parents=True)

        version = self.config.bigscape.version

        run_bigscape(
            self.antismash_dir,
            self.bigscape_running_output_dir,
            self.config.bigscape.parameters,
            version,
        )

        if version == 1:
            for f in glob(
                str(
                    self.bigscape_running_output_dir
                    / "network_files"
                    / "*"
                    / "mix"
                    / "mix_clustering_c*.tsv"
                )
            ):
                shutil.copy(f, self.bigscape_dir)
        elif version == 2:
            shutil.copy(
                self.bigscape_running_output_dir / "data_sqlite.db",
                self.bigscape_dir,
            )
        else:
            raise ValueError(f"Invalid BiG-SCAPE version: {version}")

    def arrange_strain_mappings(self) -> None:
        """Arrange the strain mappings file.

        For `local` mode, validate the strain mappings file.

        For `podp` mode, always generate the new strain mappings file and validate it.

        The validation checks if the strain mappings file exists and if it is a valid JSON file
        according to [STRAIN_MAPPINGS_SCHEMA][nplinker.schemas.STRAIN_MAPPINGS_SCHEMA].
        """
        if self.config.mode == "podp":
            self._generate_strain_mappings()

        self._validate_strain_mappings()

    def _validate_strain_mappings(self) -> None:
        """Validate the strain mappings file.

        The validation process includes:

        - Check if the strain mappings file exists.
        - Check if the strain mappings file is a valid JSON file according to
            [STRAIN_MAPPINGS_SCHEMA][nplinker.schemas.STRAIN_MAPPINGS_SCHEMA].

        Raises:
            FileNotFoundError: If the strain mappings file is not found.
            ValidationError: If the strain mappings file is not a valid JSON file according to the
                schema.
        """
        strain_mappings_file = self.root_dir / defaults.STRAIN_MAPPINGS_FILENAME

        if not strain_mappings_file.exists():
            raise FileNotFoundError(f"Strain mappings file not found at {strain_mappings_file}")

        with open(strain_mappings_file, "r") as f:
            json_data = json.load(f)
        # Will raise ValidationError if the JSON file is invalid
        validate(instance=json_data, schema=STRAIN_MAPPINGS_SCHEMA)

    def _generate_strain_mappings(self) -> None:
        """Generate the strain mappings file for the PODP mode."""
        podp_json_file = self.downloads_dir / f"paired_datarecord_{self.config.podp_id}.json"
        genome_status_json_file = self.downloads_dir / defaults.GENOME_STATUS_FILENAME
        genome_bgc_mappings_file = self.antismash_dir / defaults.GENOME_BGC_MAPPINGS_FILENAME
        gnps_file_mapping_file = self.gnps_file_mappings_file
        strain_mappings_file = self.root_dir / defaults.STRAIN_MAPPINGS_FILENAME

        # generate the genome_bgc_mappings_file
        generate_mappings_genome_id_bgc_id(self.antismash_dir)
        # generate the strain_mappings_file
        podp_generate_strain_mappings(
            podp_json_file,
            genome_status_json_file,
            genome_bgc_mappings_file,
            gnps_file_mapping_file,
            strain_mappings_file,
        )

    def arrange_strains_selected(self) -> None:
        """Arrange the strains selected file.

        If the file exists, validate it according to the schema defined in `user_strains.json`.
        """
        strains_selected_file = self.root_dir / defaults.STRAINS_SELECTED_FILENAME
        if strains_selected_file.exists():
            with open(strains_selected_file, "r") as f:
                json_data = json.load(f)
            validate(instance=json_data, schema=USER_STRAINS_SCHEMA)


def validate_gnps(gnps_dir: str | PathLike) -> None:
    """Validate the GNPS data directory and its contents.

    The GNPS data directory must contain the following files:

    - `file_mappings.tsv` or `file_mappings.csv`
    - `spectra.mgf`
    - `molecular_families.tsv`
    - `annotations.tsv`

    Args:
        gnps_dir: Path to the GNPS data directory.

    Raises:
        FileNotFoundError: If the GNPS data directory is not found or any of the required files
            is not found.
        ValueError: If both file_mappings.tsv and file_mapping.csv are found.
    """
    gnps_dir = Path(gnps_dir)
    if not gnps_dir.exists():
        raise FileNotFoundError(f"GNPS data directory not found at {gnps_dir}")

    file_mappings_tsv = gnps_dir / defaults.GNPS_FILE_MAPPINGS_TSV
    file_mappings_csv = gnps_dir / defaults.GNPS_FILE_MAPPINGS_CSV
    if file_mappings_tsv.exists() and file_mappings_csv.exists():
        raise ValueError(
            f"Both {file_mappings_tsv.name} and {file_mappings_csv.name} found in GNPS directory "
            f"{gnps_dir}, only one is allowed."
        )
    elif not file_mappings_tsv.exists() and not file_mappings_csv.exists():
        raise FileNotFoundError(
            f"Neither {file_mappings_tsv.name} nor {file_mappings_csv.name} found in GNPS directory"
            f" {gnps_dir}"
        )

    required_files = [
        gnps_dir / defaults.GNPS_SPECTRA_FILENAME,
        gnps_dir / defaults.GNPS_MOLECULAR_FAMILY_FILENAME,
        gnps_dir / defaults.GNPS_ANNOTATIONS_FILENAME,
    ]
    list_not_found = [f.name for f in required_files if not f.exists()]
    if list_not_found:
        raise FileNotFoundError(
            f"Files not found in GNPS directory {gnps_dir}: ', '.join({list_not_found})"
        )


def validate_antismash(antismash_dir: str | PathLike) -> None:
    """Validate the antiSMASH data directory and its contents.

    The validation only checks the structure of the antiSMASH data directory and file names.
    It does not check

    - the content of the BGC files
    - the consistency between the antiSMASH data and the PODP project JSON file for the `podp` mode

    The antiSMASH data directory must exist and contain at least one sub-directory. The name of the
    sub-directories must not contain any space. Each sub-directory must contain at least one BGC
    file (with the suffix `.region???.gbk` where `???` is the region number).

    Args:
        antismash_dir: Path to the antiSMASH data directory.

    Raises:
        FileNotFoundError: If the antiSMASH data directory is not found, or no sub-directories
            are found in the antiSMASH data directory, or no BGC files are found in any
            sub-directory.
        ValueError: If any sub-directory name contains a space.
    """
    antismash_dir = Path(antismash_dir)
    if not antismash_dir.exists():
        raise FileNotFoundError(f"antiSMASH data directory not found at {antismash_dir}")

    sub_dirs = list_dirs(antismash_dir)
    if not sub_dirs:
        raise FileNotFoundError(
            "No BGC directories found in antiSMASH data directory {antismash_dir}"
        )

    for sub_dir in sub_dirs:
        dir_name = Path(sub_dir).name
        if " " in dir_name:
            raise ValueError(
                f"antiSMASH sub-directory name {dir_name} contains space, which is not allowed"
            )

        gbk_files = list_files(sub_dir, suffix=".gbk", keep_parent=False)
        bgc_files = fnmatch.filter(gbk_files, "*.region???.gbk")
        if not bgc_files:
            raise FileNotFoundError(f"No BGC files found in antiSMASH sub-directory {sub_dir}")


def validate_bigscape(bigscape_dir: str | PathLike, cutoff: str) -> None:
    """Validate the BiG-SCAPE data directory and its contents.

    The BiG-SCAPE data directory must exist and contain the clustering file
    `mix_clustering_c{self.config.bigscape.cutoff}.tsv` where `{self.config.bigscape.cutoff}` is the
    bigscape cutoff value set in the config file.

    Alternatively, the directory can contain the BiG-SCAPE database file generated by BiG-SCAPE v2.
    At the moment, all the family assignments in the database will be used, so this database should
    contain results from a single run with the desired cutoff.

    Args:
        bigscape_dir: Path to the BiG-SCAPE data directory.
        cutoff: The BiG-SCAPE cutoff value.

    Raises:
        FileNotFoundError: If the BiG-SCAPE data directory or the clustering file is not found.
    """
    bigscape_dir = Path(bigscape_dir)
    if not bigscape_dir.exists():
        raise FileNotFoundError(f"BiG-SCAPE data directory not found at {bigscape_dir}")

    clustering_file = bigscape_dir / f"mix_clustering_c{cutoff}.tsv"
    database_file = bigscape_dir / "data_sqlite.db"
    if not clustering_file.exists() and not database_file.exists():
        raise FileNotFoundError(f"BiG-SCAPE data not found in {clustering_file} or {database_file}")



from .bigscape_loader import BigscapeGCFLoader
from .bigscape_loader import BigscapeV2GCFLoader
from .runbigscape import run_bigscape


__all__ = ["BigscapeGCFLoader", "BigscapeV2GCFLoader", "run_bigscape"]



import logging
import time
from nplinker.genomics import BGC
from nplinker.genomics import GCF
from nplinker.metabolomics import Spectrum
from nplinker.strain import StrainCollection
from .abc import ScoringBase
from .link_graph import LinkGraph
from .score import Score
from .scoring_method import ScoringMethod


logger = logging.getLogger(__name__)


class NPClassScoring(ScoringBase):
    name = ScoringMethod.NPLCLASS.value

    def __init__(self, npl):
        super().__init__(npl)
        self.cutoff = 0.25
        self.method_options = npl.chem_classes.class_predict_options
        self.method = self.method_options[0]
        self.num_results = 1  # how many scores do you want for each link
        # take care what targets are as this method can link both bgc/gcf to
        # both spectrum/gcf
        self.equal_targets = False  # if obj GCF, target is spec not MF
        self.both_targets = False  # if obj GCF, target both spec and MF
        # filter out spectra without a score due to missing spectrum classes
        self.filter_missing_scores = False
        self._target_no_scores = set()

    def _npclass_score(self, obj, target, method="mix", obj_classes=None, target_classes=None):
        """Return sorted class link scores for scoring obj and target.

        The input objects can be any mix of the following NPLinker types:
            - BGC
            - GCF
            - Spectrum
            - MolecularFamily

        Classes will be retrieved from NPLinker object with _get_gen_classes
        and _get_met_classes or the classes can be predetermined and given as
        input with the function, this should correspond to obj and
        target (if obj is BGC then  obj_classes should be classes for the BGC).
        Can be:
            - BGC/GCF classes: {'as_classes': list(str), 'bigscape_class': str}
            - Spectrum/mf classes: tuple of (classes, class_names_indices),
            where classes is a list of list of tuples/None, where each
            tuple is a class and a score (str, float), and class_names_indices
            is a list of ints that relate to the name of a class ontology lvl

        Args:
            obj: one of the possible input objects
            target: one of the possible input objects
            method: str, which classification method to use for spectra. default is 'main'
                options:
                    -'main': use the main method - currently canopus
                    -'mix': use main method first, when no main classes present, use the others
                    if present
                    -'canopus': use only canopus class predictions
                    -'molnetenhancer': use only molnetenhancer class predictions
            obj_classes: default - None, or classes for one of the input types
            target_classes: default - None, or classes for one of the input types

        Returns:
            List of tuple of
            (score, obj_class_lvl, target_class_lvl, obj_class, target_class)
            (float, str, str, str, str)
            List will be empty if either BGC or spectrum classes are missing
        """
        # todo: make subroutines
        # assess what is obj and target
        spec_like = obj
        bgc_like = target
        spec_like_classes_tup = obj_classes
        bgc_like_classes_dict = target_classes
        bgc_to_spec = False
        if isinstance(obj, BGC) or isinstance(obj, GCF):
            bgc_like = obj
            spec_like = target
            bgc_like_classes_dict = obj_classes
            spec_like_classes_tup = target_classes
            bgc_to_spec = True

        # assess method - move to get_links?
        assert method in self.method_options, (
            f"NPClass method should be one of method options: {self.method_options}, if your method is not "
            + "in the options check if the class predictions (canopus, etc.) are loaded correctly"
        )

        # gather correct classes if not provided, dict for bgcs, tup for spec
        if not bgc_like_classes_dict:
            bgc_like_classes_dict = self._get_gen_classes(bgc_like)
        if not spec_like_classes_tup:
            spec_like_classes_tup = self._get_met_classes(spec_like, method)
        # unpack spec_like classes - both are lists
        spec_like_classes, spec_like_classes_names_inds = spec_like_classes_tup

        scores = []  # this will be returned if one of the class sides is absent
        std_score = 0  # if link not recorded in scores (mibig) return this score
        # loop through classes that are possible to link (names in class_match object)
        for bgc_class_name in self.npl.class_matches.bgc_class_names:
            if bgc_class_name == "mibig_classes":
                # treat specially as bigscape class needs to be translated to mibig class
                bigscape_class = bgc_like_classes_dict["bigscape_class"]
                # convert bigscape class to mibig class
                bgc_like_classes = [
                    self.npl.class_matches.bigscape_mibig_conversion.get(bigscape_class)
                ]
            else:
                bgc_like_classes = bgc_like_classes_dict.get(bgc_class_name)
            if bgc_like_classes and spec_like_classes:  # check for classes from both sides
                for bgc_class in bgc_like_classes:
                    for chem_class_name in self.npl.class_matches.chem_class_names:
                        # does info exist for this spectrum class level, return index for class level
                        spec_class_level_i = spec_like_classes_names_inds.get(chem_class_name)
                        if spec_class_level_i:
                            spec_class_tup = spec_like_classes[spec_class_level_i]
                            if spec_class_tup:  # if there is a class at this lvl
                                # is a tuple of (name, score) so take [0]
                                spec_class = spec_class_tup[0]
                                # determine direction of scoring: BGC -> spectrum
                                if bgc_to_spec:
                                    score = (
                                        self.npl.class_matches.class_matches[bgc_class_name][
                                            chem_class_name
                                        ]
                                        .get(bgc_class, {})
                                        .get(spec_class, std_score)
                                    )
                                    result_tuple = (
                                        score,
                                        bgc_class_name,
                                        chem_class_name,
                                        bgc_class,
                                        spec_class,
                                    )
                                else:  # spectrum -> BGC
                                    score = (
                                        self.npl.class_matches.class_matches[chem_class_name][
                                            bgc_class_name
                                        ]
                                        .get(spec_class, {})
                                        .get(bgc_class, std_score)
                                    )
                                    result_tuple = (
                                        score,
                                        chem_class_name,
                                        bgc_class_name,
                                        spec_class,
                                        bgc_class,
                                    )
                                scores.append(result_tuple)
        return sorted(scores, reverse=True)

    def _get_targets(self, test_id):
        """Get the targets based upon instance of test_id, returns list of targets.

        Args:
            test_id: one of the NPLinker objects: BGC, GCF, Spectrum, mf
        Returns:
            List of one or more of one of the NPLinker objects
        """
        if isinstance(test_id, BGC) or isinstance(test_id, GCF):  # genome side
            return self._get_targets_genomics(test_id)
        else:  # metabolome side
            return self._get_targets_metabolomics(test_id)

    def _get_targets_metabolomics(self, test_id):
        if self.both_targets:
            targets = self.npl.bgcs + self.npl.gcfs
        elif isinstance(test_id, Spectrum):
            if self.equal_targets:
                targets = self.npl.bgcs
            else:
                targets = self.npl.gcfs
        else:  # obj are mf
            if self.equal_targets:
                targets = self.npl.gcfs
            else:
                targets = self.npl.bgcs
        return targets

    def _get_targets_genomics(self, test_id):
        if self.both_targets:  # no matter BGC or GCF take both spec and MF
            targets = self.npl.spectra + self.npl.mfs
        elif isinstance(test_id, BGC):  # obj are BGC
            if self.equal_targets:  # take
                targets = self.npl.spectra
            else:
                targets = self.npl.mfs
        else:  # obj are GCF
            if self.equal_targets:
                targets = self.npl.mfs
            else:
                targets = self.npl.spectra
        return targets

    def _get_gen_classes(self, bgc_like, gcf_as_cutoff=0.5):
        """Get classes for genomics objects.

        Args:
            bgc_like: BGC or GCF object from NPLinker input objects
            gcf_as_cutoff: float - if GCF, get antismash classes with
                class_matches.get_gcf_as_classes if the class occurs in more
                than this fraction of the GCF, default = 0.5
        Returns:
            Genome-based classes as a dict with antismash and bigscpae classes
            {'as_classes': list(str), 'bigscape_class': str}
        """
        # assess if bgc or gcf
        is_bgc = isinstance(bgc_like, BGC)
        if is_bgc:
            # get parent gcf for bgc
            bgc_like_gcf = [
                gcf for gcf in self.npl.gcfs if bgc_like.id in [b.id for b in gcf.bgcs]
            ][0]
            # gather AS classes and convert to names in scoring dict
            as_classes = self.npl.class_matches.convert_as_classes(
                bgc_like.product_prediction.split(".")
            )
            bgc_like_classes_dict = {
                "bigscape_class": bgc_like_gcf.bigscape_class,
                # str - always one bigscape class right?
                "as_classes": as_classes,
            }  # list(str)
        else:
            as_classes = self.npl.class_matches.convert_as_classes(
                self.npl.class_matches.get_gcf_as_classes(bgc_like, gcf_as_cutoff)
            )
            bgc_like_classes_dict = {
                "bigscape_class": bgc_like.bigscape_class,
                # str - always one bigscape class right?
                "as_classes": as_classes,
            }  # list(str)
        return bgc_like_classes_dict

    def _get_met_classes(self, spec_like, method="mix"):
        """Get chemical classes for a Spectrum or mf based on method.

        Args:
            spec_like: Spectrum or mf, one of the NPLinker input types
            method: str, one of the appropriate methods for chemical class
                predictions (mix, canopus...), default='mix'
        Returns:
            tuple of (classes, class_names_indices),
            where classes is a list of list of tuples/None, where each
            tuple is a class and a score (str, float), and class_names_indices
            is a list of ints that relate to the name of a class ontology lvl
        """
        # assess if spectrum or mf
        is_spectrum = isinstance(spec_like, Spectrum)

        # gather classes for spectra, using right method
        # choose the main method here by including it as 'main' in the method parameter
        use_canopus = (
            "main" in method or "canopus" in method or "mix" in method
        ) and "canopus" in self.method_options
        use_mne = (
            "molnetenhancer" in method or "mix" in method
        ) and "molnetenhancer" in self.method_options
        spec_like_classes, spec_like_classes_names, spec_like_classes_names_inds = (
            None,
            None,
            None,
        )
        # the order in which the classes are read, determines the priority (now: first canopus, then mne)
        if use_canopus and not spec_like_classes:
            if is_spectrum:
                # list of list of tuples/None - todo: add to spectrum object?
                # take only 'best' (first) classification per ontology level
                all_classes = self.npl.chem_classes.canopus.spectra_classes.get(spec_like.id)
                if all_classes:
                    spec_like_classes = [
                        cls_per_lvl
                        for lvl in all_classes
                        for i, cls_per_lvl in enumerate(lvl)
                        if i == 0
                    ]
                spec_like_classes_names_inds = (
                    self.npl.chem_classes.canopus.spectra_classes_names_inds
                )
            else:  # mf
                fam_id = spec_like.family.id
                if fam_id.startswith("singleton-"):  # account for singleton families
                    fam_id += f"_{spec_like.spectra[0].id}"
                all_classes = self.npl.chem_classes.canopus.mf_classes.get(fam_id)
                if all_classes:
                    spec_like_classes = [
                        cls_per_lvl
                        for lvl in all_classes
                        for i, cls_per_lvl in enumerate(lvl)
                        if i == 0
                    ]
                spec_like_classes_names_inds = self.npl.chem_classes.canopus.mf_classes_names_inds
        if use_mne and not spec_like_classes:
            # if mne or when main/canopus does not get classes
            if is_spectrum:
                spec_like_classes = self.npl.chem_classes.molnetenhancer.spectra_classes(
                    spec_like.id
                )
            else:  # mf
                fam_id = spec_like.family.id
                if fam_id.startswith("singleton"):  # account for singleton families
                    fam_id += f"_{spec_like.spectra[0].id}"
                spec_like_classes = self.npl.chem_classes.molnetenhancer.mf_classes.get(fam_id)
            # classes are same for mf and spectrum so names are irrespective of is_spectrum
            spec_like_classes_names_inds = (
                self.npl.chem_classes.molnetenhancer.spectra_classes_names_inds
            )
        return spec_like_classes, spec_like_classes_names_inds

    @classmethod
    def setup(cls, npl):
        """Perform any one-off initialisation required (will only be called once)."""
        logger.info("Set up NPClassScore scoring")
        met_options = npl.chem_classes.class_predict_options
        logger.info(f"Please choose one of the methods from {met_options}")
        if not met_options:
            logger.warn(
                "There are no methods available! This is probably "
                "because no class predictions (canopus, etc.) were "
                "present or they are not loaded correctly"
            )
        else:
            logger.info(f"Currently the method '{met_options[0]}' is selected")
        # todo: give info about parameters

    def get_links(self, *objects, **parameters):
        # TODO: replace some attributes with parameters
        """Given a set of objects, return link information."""
        # todo: pickle results
        logger.info("Running NPClassScore...")
        begin = time.time()
        first_obj = objects[0]
        targets = self._get_targets(first_obj)
        obj_is_gen = isinstance(first_obj, BGC) or isinstance(first_obj, GCF)

        # only get target classes once for each target here
        if obj_is_gen:  # obj is genome so get metabolome classes for target
            targets_classes = [self._get_met_classes(target, self.method) for target in targets]
        else:
            targets_classes = [self._get_gen_classes(target) for target in targets]

        # TODO: implement the computation of common strains between objects and targets
        common_strains = StrainCollection()

        logger.info(
            f"Calculating NPClassScore for {len(objects)} objects to "
            f"{len(targets)} targets ({len(common_strains)} pairwise "
            f"interactions that share at least 1 strain). This might "
            f"take a while."
        )

        lg = LinkGraph()
        results = {}
        for obj in objects:
            results[obj] = {}
            # get obj class
            if obj_is_gen:
                obj_classes = self._get_gen_classes(obj)
            else:
                obj_classes = self._get_met_classes(obj, self.method)

            for target, target_classes in zip(targets, targets_classes):
                self._create_object_link(
                    obj_is_gen,
                    common_strains,
                    lg,
                    obj,
                    obj_classes,
                    target,
                    target_classes,
                    parameters,
                )

        # info about spectra/MFs with missing scoring
        len_missing = len(self._target_no_scores)
        if len_missing > 0:
            filter_msg = "kept"
            if self.filter_missing_scores:
                filter_msg = "filtered out"
            logger.warning(
                f"{len_missing} targets have no NPClassScore "
                f"prediction due to missing class predictions and are "
                f"{filter_msg} by default. Adjust .filter_missing_scores "
                f"to change."
            )

        logger.info(f"NPClassScore completed in {time.time() - begin:.1f}s")
        return lg

    def _create_object_link(
        self, obj_is_gen, common_strains, lg, obj, obj_classes, target, target_classes, parameters
    ):
        # only consider targets that have shared strains
        common_tup = (obj, target)
        if obj_is_gen:
            common_tup = (target, obj)
        shared_strains = common_strains.get(common_tup)
        if shared_strains is not None:  # todo: fix for bgcs
            full_score = self._npclass_score(obj, target, self.method, obj_classes, target_classes)[
                : self.num_results
            ]
            try:
                npclassscore = full_score[0][0]
            except IndexError:
                # no score is found due to missing classes for spectra
                self._target_no_scores.add(target)  # keep track
                if not self.filter_missing_scores:
                    lg.add_link(obj, target, Score(self.name, full_score, parameters))
            else:
                if npclassscore > self.cutoff:
                    lg.add_link(obj, target, Score(self.name, full_score, parameters))

    def format_data(self, data):
        """Given whatever output data the method produces, return a readable string version."""
        # data or full_score is a list of tuples, here return just NPClassScore
        formatted_data = None  # default when there is no score (missing class)
        if data:
            # there is a score
            formatted_data = f"{data[0][0]:.3f}"
        return formatted_data

    def sort(self, objects, reverse=True):
        """Given a list of objects, return them sorted by link score."""
        return sorted(objects, key=lambda objlink: objlink[self], reverse=reverse)



import logging
from rich.console import Console
from rich.logging import RichHandler


def setup_logging(level: str = "INFO", file: str = "", use_console: bool = True) -> None:
    """Setup logging configuration for the ancestor logger "nplinker".

    ??? info "Usage Documentation"
        [How to setup logging][how-to-setup-logging]

    Args:
        level: The log level, use the logging module's log level constants.
            Valid levels are: `NOTSET`, `DEBUG`, `INFO`, `WARNING`, `ERROR`, `CRITICAL`.
        file: The file to write the log to.
            If the file is an empty string (by default), the log will not be written to a file.
            If the file does not exist, it will be created.
            The log will be written to the file in append mode.
        use_console: Whether to log to the console.
    """
    # Get the ancestor logger "nplinker"
    logger = logging.getLogger("nplinker")
    logger.setLevel(level)
    logger.handlers.clear()

    # File handler
    if file:
        logger.addHandler(
            RichHandler(
                console=Console(file=open(file, "a"), width=120),  # force the line width to 120
                omit_repeated_times=False,
                rich_tracebacks=True,
                tracebacks_show_locals=True,
                log_time_format="[%Y-%m-%d %X]",
            )
        )

    # Console handler
    if use_console:
        logger.addHandler(
            RichHandler(
                omit_repeated_times=False,
                rich_tracebacks=True,
                tracebacks_show_locals=True,
                log_time_format="[%Y-%m-%d %X]",
            )
        )



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re


class Formula:
    def __init__(self, formula):
        self.atom_names = ["C", "H", "N", "O", "P", "S", "Cl", "I", "Br", "Si", "F", "D"]
        self.formula = formula
        self.atoms = {}
        for atom in self.atom_names:
            self.atoms[atom] = self.get_atoms(atom)

    def equals(self, formula):
        is_equal = True
        for atom in self.atoms:
            if not self.atoms[atom] == formula.atoms[atom]:
                is_equal = False
                return is_equal
        return is_equal

    def correct_gcms_derivatives(self):
        n_silicons = self.atoms["Si"]
        self.atoms["Si"] = 0
        self.atoms["C"] -= n_silicons
        self.atoms["H"] -= 3 * n_silicons
        self.atoms["H"] += n_silicons
        self.make_string()

    def make_string(self):
        self.formula = ""
        for atom in self.atom_names:
            atom_no = self.atoms[atom]
            if atom_no == 1:
                self.formula += atom
            elif atom_no > 1:
                self.formula += atom + str(atom_no)

    def get_atoms(self, atom_name):
        # Do some regex matching to find the numbers of the important atoms
        ex = atom_name + "(?![a-z])" + r"\d*"
        m = re.search(ex, self.formula)
        if m is None:
            return 0
        else:
            ex = atom_name + "(?![a-z])" + r"(\d*)"
            m2 = re.findall(ex, self.formula)
            total = 0
            for a in m2:
                if len(a) == 0:
                    total += 1
                else:
                    total += int(a)
            return total

    def compute_exact_mass(self):
        masses = {
            "C": 12.00000000000,
            "H": 1.00782503214,
            "O": 15.99491462210,
            "N": 14.00307400524,
            "P": 30.97376151200,
            "S": 31.97207069000,
            "Cl": 34.96885271000,
            "I": 126.904468,
            "Br": 78.9183376,
            "Si": 27.9769265327,
            "F": 18.99840320500,
            "D": 2.01410177800,
        }
        exact_mass = 0.0
        for a in self.atoms:
            exact_mass += masses[a] * self.atoms[a]
        return exact_mass

    def __str__(self):
        fstring = ""
        for atom in self.atom_names:
            if self.atoms[atom] > 0:
                fstring += atom
            if self.atoms[atom] > 1:
                fstring += str(self.atoms[atom])

        return fstring

    def __repr__(self):
        return self.__str__()



from __future__ import annotations
import csv
from os import PathLike
from nplinker.metabolomics.abc import MolecularFamilyLoaderBase
from nplinker.utils import is_file_format
from ..molecular_family import MolecularFamily


class GNPSMolecularFamilyLoader(MolecularFamilyLoaderBase):
    """Load molecular families from GNPS data.

    ??? info "Concept"
        [GNPS data][gnps-data]

    The molecular family file is from GNPS output archive, as described below
    for each GNPS workflow type:

    1. METABOLOMICS-SNETS
        - networkedges_selfloop/*.pairsinfo
    2. METABOLOMICS-SNETS-V2
        - networkedges_selfloop/*.selfloop
    3. FEATURE-BASED-MOLECULAR-NETWORKING
        - networkedges_selfloop/*.selfloop

    The `ComponentIndex` column in the GNPS molecular family file is treated
    as family id.

    But for molecular families that have only one member (i.e. spectrum),
    named singleton molecular families, their files have the same value of
    `-1` in the `ComponentIndex` column. To make the family id unique,the
    spectrum id plus a prefix `singleton-` is used as the family id of
    singleton molecular families.
    """

    def __init__(self, file: str | PathLike) -> None:
        """Initialize the GNPSMolecularFamilyLoader.

        Args:
            file: Path to the GNPS molecular family file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.

        Examples:
            >>> loader = GNPSMolecularFamilyLoader("gnps_molecular_families.tsv")
            >>> print(loader.families)
            [<MolecularFamily 1>, <MolecularFamily 2>, ...]
            >>> print(loader.families[0].spectra_ids)
            {'1', '3', '7', ...}
        """
        self._mfs: list[MolecularFamily] = []
        self._file = file

        self._validate()
        self._load()

    def get_mfs(self, keep_singleton: bool = False) -> list[MolecularFamily]:
        """Get MolecularFamily objects.

        Args:
            keep_singleton: True to keep singleton molecular families. A
                singleton molecular family is a molecular family that contains
                only one spectrum.

        Returns:
            A list of MolecularFamily objects with their spectra ids.
        """
        mfs = self._mfs
        if not keep_singleton:
            mfs = [mf for mf in mfs if not mf.is_singleton()]
        return mfs

    def _validate(self) -> None:
        """Validate the GNPS molecular family file."""
        # validate file format
        if not is_file_format(self._file, "tsv"):
            raise ValueError(
                f"Invalid GNPS molecular family file '{self._file}'. " f"Expected a '.tsv' file."
            )
        # validate required columns against the header
        required_columns = ["CLUSTERID1", "CLUSTERID2", "ComponentIndex"]
        with open(self._file, mode="rt") as f:
            header = f.readline()
            for k in required_columns:
                if k not in header:
                    raise ValueError(
                        f"Invalid GNPS molecular famliy file '{self._file}'. "
                        f"Expected a header line with '{k}' column, "
                        f"but got '{header}'."
                    )

    def _load(self) -> None:
        """Load molecular families from GNPS output file.

        Molecular families are loaded as a list of MolecularFamily objects. Each
        MolecularFamily object contains a set of spectra ids that belong to this
        family.
        """
        # load molecular families to dict
        family_dict = {}
        with open(self._file, mode="rt", encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                spec1_id = row["CLUSTERID1"]
                spec2_id = row["CLUSTERID2"]
                mf_id = row["ComponentIndex"]
                if mf_id not in family_dict:
                    family_dict[mf_id] = set([spec1_id, spec2_id])
                else:
                    family_dict[mf_id].add(spec1_id)
                    family_dict[mf_id].add(spec2_id)
        # convert dict to list of MolecularFamily objects
        for mf_id, spectra_ids in family_dict.items():
            if mf_id == "-1":  # "-1" is from GNPS, it means the singleton molecular family
                for spectrum_id in spectra_ids:
                    # family id must be unique, so using "singleton-" + spectrum id as family id
                    family = MolecularFamily("singleton-" + str(spectrum_id))
                    family.spectra_ids = set([spectrum_id])
                    self._mfs.append(family)
            else:
                # for regular molecular families, use the value of "ComponentIndex" as family id
                family = MolecularFamily(mf_id)
                family.spectra_ids = spectra_ids
                self._mfs.append(family)



from .mibig_downloader import download_and_extract_mibig_metadata
from .mibig_loader import MibigLoader
from .mibig_loader import parse_bgc_metadata_json
from .mibig_metadata import MibigMetadata


__all__ = [
    "download_and_extract_mibig_metadata",
    "MibigLoader",
    "MibigMetadata",
    "parse_bgc_metadata_json",
]



import json
from pathlib import Path
from jsonschema import Draft7Validator


__all__ = [
    "GENOME_STATUS_SCHEMA",
    "GENOME_BGC_MAPPINGS_SCHEMA",
    "STRAIN_MAPPINGS_SCHEMA",
    "PODP_ADAPTED_SCHEMA",
    "USER_STRAINS_SCHEMA",
    "validate_podp_json",
]

_schema_dir = Path(__file__).parent

with open(_schema_dir / "genome_status_schema.json", "r") as f:
    GENOME_STATUS_SCHEMA = json.load(f)
    """Schema for the [genome status JSON file][nplinker.defaults.GENOME_STATUS_FILENAME].

    ??? info "Schema Content: `genome_status_schema.json`"
        ```json
        --8<-- "src/nplinker/schemas/genome_status_schema.json"
        ```
    """

with open(_schema_dir / "genome_bgc_mappings_schema.json", "r") as f:
    GENOME_BGC_MAPPINGS_SCHEMA = json.load(f)
    """Schema for [genome BGC mappings JSON file][nplinker.defaults.GENOME_BGC_MAPPINGS_FILENAME].

    ??? info "Schema Content: `genome_bgc_mappings_schema.json`"
        ```json
        --8<-- "src/nplinker/schemas/genome_bgc_mappings_schema.json"
        ```
    """

with open(_schema_dir / "strain_mappings_schema.json", "r") as f:
    STRAIN_MAPPINGS_SCHEMA = json.load(f)
    """Schema for [strain mappings JSON file][nplinker.defaults.STRAIN_MAPPINGS_FILENAME].

    ??? info "Schema Content: `strain_mappings_schema.json`"
        ```json
        --8<-- "src/nplinker/schemas/strain_mappings_schema.json"
        ```
    """

with open(_schema_dir / "user_strains.json", "r") as f:
    USER_STRAINS_SCHEMA = json.load(f)
    """Schema for [user strains JSON file][nplinker.defaults.STRAINS_SELECTED_FILENAME].

    ??? info "Schema Content: `user_strains.json`"
        ```json
        --8<-- "src/nplinker/schemas/user_strains.json"
        ```
    """

with open(_schema_dir / "podp_adapted_schema.json", "r") as f:
    PODP_ADAPTED_SCHEMA = json.load(f)
    """Schema for PODP JSON file.

    The PODP JSON file is the project JSON file downloaded from PODP platform.
    For example, for PODP project MSV000079284, its JSON file is
    https://pairedomicsdata.bioinformatics.nl/api/projects/4b29ddc3-26d0-40d7-80c5-44fb6631dbf9.4.

    ??? info "Schema Content: `podp_adapted_schema.json`"
        ```json
        --8<-- "src/nplinker/schemas/podp_adapted_schema.json"
        ```
    """


def validate_podp_json(json_data: dict) -> None:
    """Validate JSON data against the PODP JSON schema.

    All validation error messages are collected and raised as a single
    ValueError.

    Args:
        json_data: The JSON data to validate.

    Raises:
        ValueError: If the JSON data does not match the schema.

    Examples:
        Download PODP JSON file for project MSV000079284 from
        https://pairedomicsdata.bioinformatics.nl/api/projects/4b29ddc3-26d0-40d7-80c5-44fb6631dbf9.4
        and save it as `podp_project.json`.

        Validate it:
        >>> with open(podp_project.json, "r") as f:
        ...     json_data = json.load(f)
        >>> validate_podp_json(json_data)
    """
    validator = Draft7Validator(PODP_ADAPTED_SCHEMA)
    errors = sorted(validator.iter_errors(json_data), key=lambda e: e.path)
    if errors:
        error_messages = [f"{e.json_path}: {e.message}" for e in errors]
        raise ValueError(
            "Not match PODP adapted schema, here are the detailed error:\n  - "
            + "\n  - ".join(error_messages)
        )



from __future__ import annotations
import json
from os import PathLike


class MibigMetadata:
    """Class to model the BGC metadata/annotations defined in MIBiG.

    MIBiG is a specification of BGC metadata and use JSON schema to
    represent BGC metadata. More details see:
    https://mibig.secondarymetabolites.org/download.

    This class supports MIBiG version 1.0 to 4.0.
    """

    def __init__(self, file: str | PathLike) -> None:
        """Initialize the MIBiG metadata object.

        Args:
            file: Path to the json file of MIBiG BGC metadata

        Examples:
            >>> metadata = MibigMetadata("/data/BGC0000001.json")
        """
        self.file = str(file)
        with open(self.file, "rb") as f:
            self.metadata = json.load(f)

        self._mibig_accession: str
        self._biosyn_class: tuple[str]
        self._parse_metadata()

    @property
    def mibig_accession(self) -> str:
        """Get the value of metadata item 'mibig_accession'."""
        return self._mibig_accession

    @property
    def biosyn_class(self) -> tuple[str]:
        """Get the value of metadata item 'biosyn_class'.

        The 'biosyn_class' is biosynthetic class(es) defined by MIBiG.

        Before version 4.0 of MIBiG, it defines 6 major biosynthetic classes,
        including `NRP`, `Polyketide`, `RiPP`, `Terpene`, `Saccharide`
        and `Alkaloid`.

        Starting from version 4.0, MIBiG defines 5 major biosynthetic classes,
        including `PKS`, `NRPS`, `Ribosomal`, `Terpene` and `Saccharide`.

        The mapping between the old and new classes is as follows:

        - `NRP` -> `NRPS`
        - `Polyketide` -> `PKS`
        - `RiPP` -> `Ribosomal`
        - `Terpene` -> `Terpene`
        - `Saccharide` -> `Saccharide`
        - `Alkaloid` -> `Other`

        Note that natural products that do not fit into any of the above
        biosynthetic classes fall under the category `Other`.
        """
        return self._biosyn_class

    def _parse_metadata(self) -> None:
        """Parse metadata to get 'mibig_accession' and 'biosyn_class' values."""
        if "general_params" in self.metadata:  # version ≤1.4
            self._mibig_accession = self.metadata["general_params"]["mibig_accession"]
            self._biosyn_class = tuple(self.metadata["general_params"]["biosyn_class"])
        elif "cluster" in self.metadata:  # version ≥2.0 and <4.0
            self._mibig_accession = self.metadata["cluster"]["mibig_accession"]
            self._biosyn_class = tuple(self.metadata["cluster"]["biosyn_class"])
        elif "version" in self.metadata:  # version≥4.0
            self._mibig_accession = self.metadata["accession"]
            self._biosyn_class = tuple(i["class"] for i in self.metadata["biosynthesis"]["classes"])



from __future__ import annotations
import json
import logging
from collections.abc import Iterator
from os import PathLike
from jsonschema import validate
from nplinker.schemas import STRAIN_MAPPINGS_SCHEMA
from .strain import Strain


logger = logging.getLogger(__name__)


class StrainCollection:
    """A collection of `Strain` objects."""

    def __init__(self) -> None:
        # the order of strains is needed for scoring part, so use a list
        self._strains: list[Strain] = []
        self._strain_dict_name: dict[str, list[Strain]] = {}

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if len(self) > 20:
            return f"StrainCollection(n={len(self)})"

        return f"StrainCollection(n={len(self)}) [" + ",".join(s.id for s in self._strains) + "]"

    def __len__(self) -> int:
        return len(self._strains)

    def __eq__(self, other) -> bool:
        if isinstance(other, StrainCollection):
            return (
                self._strains == other._strains
                and self._strain_dict_name == other._strain_dict_name
            )
        return NotImplemented

    def __add__(self, other) -> StrainCollection:
        if isinstance(other, StrainCollection):
            sc = StrainCollection()
            for strain in self._strains:
                sc.add(strain)
            for strain in other._strains:
                sc.add(strain)
            return sc
        return NotImplemented

    def __contains__(self, item: Strain) -> bool:
        """Check if the strain collection contains the given Strain object."""
        if isinstance(item, Strain):
            return item.id in self._strain_dict_name
        raise TypeError(f"Expected Strain, got {type(item)}")

    def __iter__(self) -> Iterator[Strain]:
        return iter(self._strains)

    def add(self, strain: Strain) -> None:
        """Add strain to the collection.

        If the strain already exists, merge the aliases.

        Args:
            strain: The strain to add.
        """
        if strain in self._strains:
            # only one strain object per id
            strain_ref = self._strain_dict_name[strain.id][0]
            new_aliases = [alias for alias in strain.aliases if alias not in strain_ref.aliases]
            for alias in new_aliases:
                strain_ref.add_alias(alias)
                if alias not in self._strain_dict_name:
                    self._strain_dict_name[alias] = [strain_ref]
                else:
                    self._strain_dict_name[alias].append(strain_ref)
        else:
            self._strains.append(strain)
            for name in strain.names:
                if name not in self._strain_dict_name:
                    self._strain_dict_name[name] = [strain]
                else:
                    self._strain_dict_name[name].append(strain)

    def remove(self, strain: Strain) -> None:
        """Remove a strain from the collection.

        It removes the given strain object from the collection by strain id.
        If the strain id is not found, raise `ValueError`.

        Args:
            strain: The strain to remove.

        Raises:
            ValueError: If the strain is not found in the collection.
        """
        if strain in self._strains:
            self._strains.remove(strain)
            # only one strain object per id
            strain_ref = self._strain_dict_name[strain.id][0]
            for name in strain_ref.names:
                if name in self._strain_dict_name:
                    new_strain_list = [s for s in self._strain_dict_name[name] if s.id != strain.id]
                    if not new_strain_list:
                        del self._strain_dict_name[name]
                    else:
                        self._strain_dict_name[name] = new_strain_list
        else:
            raise ValueError(f"Strain {strain} not found in the strain collection.")

    def filter(self, strain_set: set[Strain]):
        """Remove all strains that are not in `strain_set` from the strain collection.

        Args:
            strain_set: Set of strains to keep.
        """
        # note that we need to copy the list of strains, as we are modifying it
        for strain in self._strains.copy():
            if strain not in strain_set:
                self.remove(strain)

    def intersection(self, other: StrainCollection) -> StrainCollection:
        """Get the intersection of two strain collections.

        Args:
            other: The other strain collection to compare.

        Returns:
            StrainCollection object containing the strains that are in both collections.
        """
        intersection = StrainCollection()
        for strain in self:
            if strain in other:
                intersection.add(strain)
        return intersection

    def has_name(self, name: str) -> bool:
        """Check if the strain collection contains the given strain name (id or alias).

        Args:
            name: Strain name (id or alias) to check.

        Returns:
            True if the strain name is in the collection, False otherwise.
        """
        return name in self._strain_dict_name

    def lookup(self, name: str) -> list[Strain]:
        """Lookup a strain by name (id or alias).

        Args:
            name: Strain name (id or alias) to lookup.

        Returns:
            List of Strain objects with the given name.

        Raises:
            ValueError: If the strain name is not found.
        """
        if name in self._strain_dict_name:
            return self._strain_dict_name[name]
        raise ValueError(f"Strain {name} not found in the strain collection.")

    @staticmethod
    def read_json(file: str | PathLike) -> StrainCollection:
        """Read a strain mappings JSON file and return a `StrainCollection` object.

        Args:
            file: Path to the strain mappings JSON file.

        Returns:
            `StrainCollection` object.
        """
        with open(file, "r") as f:
            json_data = json.load(f)

        # validate json data
        validate(instance=json_data, schema=STRAIN_MAPPINGS_SCHEMA)

        strain_collection = StrainCollection()
        for data in json_data["strain_mappings"]:
            strain = Strain(data["strain_id"])
            for alias in data["strain_alias"]:
                strain.add_alias(alias)
            strain_collection.add(strain)
        return strain_collection

    def to_json(self, file: str | PathLike | None = None) -> str | None:
        """Convert the `StrainCollection` object to a JSON string.

        Args:
            file: Path to output JSON file. If None, return the JSON string instead.

        Returns:
            If input `file` is None, return the JSON string. Otherwise, write the JSON string to the given
            file.
        """
        data_list = [
            {"strain_id": strain.id, "strain_alias": list(strain.aliases)} for strain in self
        ]
        json_data = {"strain_mappings": data_list, "version": "1.0"}

        # validate json data
        validate(instance=json_data, schema=STRAIN_MAPPINGS_SCHEMA)

        if file is not None:
            with open(file, "w") as f:
                json.dump(json_data, f)
            return None
        return json.dumps(json_data)



from __future__ import annotations
import logging
from typing import TYPE_CHECKING
from deprecated import deprecated
from nplinker.strain import Strain
from .aa_pred import predict_aa


if TYPE_CHECKING:
    from .gcf import GCF

logger = logging.getLogger(__name__)


class BGC:
    """Class to model BGC (biosynthetic gene cluster) data.

    BGC data include both annotations and sequence data. This class is
    mainly designed to model the annotations or metadata.

    The raw BGC data is stored in GenBank format (.gbk). Additional
    [GenBank features](https://www.insdc.org/submitting-standards/feature-table/)
    could be added to the GenBank file to annotate
    BGCs, e.g. antiSMASH has some self-defined features (like `region`) in
    its output GenBank files.

    The annotations of BGC can be stored in JSON format, which is defined
    and used by [MIBiG](https://mibig.secondarymetabolites.org/).

    Attributes:
        id: BGC identifier, e.g. MIBiG accession, GenBank accession.
        product_prediction: A tuple of (predicted) natural
            products or product classes of the BGC.
            For antiSMASH's GenBank data, the feature `region /product`
            gives product information.
            For MIBiG metadata, its biosynthetic class provides such info.
        mibig_bgc_class: A tuple of MIBiG biosynthetic classes to which the BGC belongs.
            Defaults to None, which means the class is unknown.

            MIBiG defines 6 major biosynthetic classes for natural products,
            including `NRP`, `Polyketide`, `RiPP`, `Terpene`, `Saccharide`
            and `Alkaloid`. Note that natural products created by the other
            biosynthetic mechanisms fall under the category `Other`. For more details
            see [the paper](https://doi.org/10.1186/s40793-018-0318-y).
        description: Brief description of the BGC.
            Defaults to None.
        smiles: A tuple of SMILES formulas of the BGC's
            products.
            Defaults to None.
        antismash_file: The path to the antiSMASH GenBank file.
            Defaults to None.
        antismash_id: Identifier of the antiSMASH BGC, referring
            to the feature `VERSION` of GenBank file.
            Defaults to None.
        antismash_region: AntiSMASH BGC region number, referring
            to the feature `region` of GenBank file.
            Defaults to None.
        parents: The set of GCFs that contain the BGC.
        strain: The strain of the BGC.
    """

    def __init__(self, id: str, /, *product_prediction: str):
        """Initialize the BGC object.

        Args:
            id: BGC identifier, e.g. MIBiG accession, GenBank accession.
            product_prediction: BGC's (predicted) natural products or product classes.

        Examples:
            >>> bgc = BGC("Unique_BGC_ID", "Polyketide", "NRP")
            >>> bgc.id
            'Unique_BGC_ID'
            >>> bgc.product_prediction
            ('Polyketide', 'NRP')
            >>> bgc.is_mibig()
            False
        """
        # BGC metadata
        self.id = id
        self.product_prediction = product_prediction

        self.mibig_bgc_class: tuple[str] | None = None
        self.description: str | None = None
        self.smiles: tuple[str] | None = None

        # antismash related attributes
        self.antismash_file: str | None = None
        self.antismash_id: str | None = None  # version in .gbk, id in SeqRecord
        self.antismash_region: int | None = None  # antismash region number

        # other attributes
        self.parents: set[GCF] = set()
        self._strain: Strain | None = None

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "{}(id={}, strain={}, asid={}, region={})".format(
            self.__class__.__name__,
            self.id,
            self.strain,
            self.antismash_id,
            self.antismash_region,
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, BGC):
            return self.id == other.id and self.product_prediction == other.product_prediction
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.id, self.product_prediction))

    def __reduce__(self) -> tuple:
        """Reduce function for pickling."""
        return (self.__class__, (self.id, *self.product_prediction), self.__dict__)

    def add_parent(self, gcf: GCF) -> None:
        """Add a parent GCF to the BGC.

        Args:
            gcf: gene cluster family
        """
        gcf.add_bgc(self)

    def detach_parent(self, gcf: GCF) -> None:
        """Remove a parent GCF."""
        gcf.detach_bgc(self)

    @property
    def strain(self) -> Strain | None:
        """Get the strain of the BGC."""
        return self._strain

    @strain.setter
    def strain(self, strain: Strain) -> None:
        self._strain = strain

    @property
    def bigscape_classes(self) -> set[str | None]:
        """Get BiG-SCAPE's BGC classes.

        BiG-SCAPE's BGC classes are similar to those defined in MiBIG but have
        more categories (7 classes), including:

        - NRPS
        - PKS-NRP_Hybrids
        - PKSI
        - PKSother
        - RiPPs
        - Saccharides
        - Terpene

        For BGC falls outside of these categories, the value is "Others".

        Default is None, which means the class is unknown.

        More details see:
        https://doi.org/10.1038%2Fs41589-019-0400-9.
        """
        return {p.bigscape_class for p in self.parents}

    def is_mibig(self) -> bool:
        """Check if the BGC is a MIBiG reference BGC or not.

        Warning:
            This method evaluates MIBiG BGC based on the pattern that MIBiG
            BGC names start with "BGC". It might give false positive result.

        Returns:
            True if it's MIBiG reference BGC
        """
        return self.id.startswith("BGC")

    # CG: why not providing whole product but only amino acid as product monomer?
    # this property is not used in NPLinker core business.
    @property
    @deprecated(version="2.0.0", reason="This method will be removed soon")
    def aa_predictions(self) -> list:
        """Amino acids as predicted monomers of product.

        Returns:
            list of dicts with key as amino acid and value as prediction
            probability.
        """
        # Load aa predictions and cache them
        self._aa_predictions = None
        if self._aa_predictions is None:
            self._aa_predictions = {}
            if self.antismash_file is not None:
                for p in predict_aa(self.antismash_file):
                    self._aa_predictions[p[0]] = p[1]
        return [self._aa_predictions]
